import torch
import torch.nn.functional as F
from torch import nn
from torch.autograd import Function

from birder.kernels.load_kernel import load_swattention

SWATTENTION_CUDA_NUM_THREADS = 128
SWATTENTION = None


def set_swattention_num_threads(num_threads: int) -> None:
    global SWATTENTION_CUDA_NUM_THREADS  # pylint: disable=global-statement
    SWATTENTION_CUDA_NUM_THREADS = num_threads


# pylint: disable=invalid-name
class SWAttention_QK_RPB(nn.Module):
    """
    TransNeXt: Robust Foveal Visual Perception for Vision Transformers: https://arxiv.org/abs/2311.17132

    Lazy-loading SWATTENTION operator.

    The custom kernel is loaded on first instantiation, not at import time.
    Falls back to pure PyTorch implementation if kernel loading fails.
    """

    def __init__(self) -> None:
        super().__init__()

        global SWATTENTION  # pylint: disable=global-statement
        if SWATTENTION is None and not torch.jit.is_tracing() and not torch.jit.is_scripting():
            SWATTENTION = load_swattention()

        self.is_available = SWATTENTION is not None

    def forward(
        self,
        kv: torch.Tensor,
        q_norm_scaled: torch.Tensor,
        relative_pos_bias_local: torch.Tensor,
        padding_mask: torch.Tensor,
        num_heads: int,
        head_dim: int,
        window_size: int,
        local_len: int,
        H: int,
        W: int,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        # Pure PyTorch
        if self.is_available is False or kv.is_cuda is False:
            return swattention_qk_rpb(
                kv,
                q_norm_scaled,
                relative_pos_bias_local,
                padding_mask,
                num_heads,
                head_dim,
                window_size,
                local_len,
                H,
                W,
            )

        # Custom kernel
        (B, N, _) = kv.size()

        # Generate unfolded keys and values and l2-normalize them
        (k_local, v_local) = kv.reshape(B, N, 2 * num_heads, head_dim).permute(0, 2, 1, 3).chunk(2, dim=1)

        # Compute local similarity
        attn_local = SWAttentionFunction_QK_RPB.apply(
            q_norm_scaled.contiguous(),
            F.normalize(k_local, dim=-1).contiguous(),
            relative_pos_bias_local,
            H,
            W,
            window_size,
        )

        return (attn_local, v_local)


# pylint: disable=invalid-name
class SWAttention_AV(nn.Module):
    """
    TransNeXt: Robust Foveal Visual Perception for Vision Transformers: https://arxiv.org/abs/2311.17132

    Lazy-loading SWATTENTION operator.

    The custom kernel is loaded on first instantiation, not at import time.
    Falls back to pure PyTorch implementation if kernel loading fails.
    """

    def __init__(self) -> None:
        super().__init__()

        global SWATTENTION  # pylint: disable=global-statement
        if SWATTENTION is None and not torch.jit.is_tracing() and not torch.jit.is_scripting():
            SWATTENTION = load_swattention()

        self.is_available = SWATTENTION is not None

    def forward(
        self,
        q_norm: torch.Tensor,
        attn_local: torch.Tensor,
        v_local: torch.Tensor,
        learnable_tokens: torch.Tensor,
        learnable_bias: torch.Tensor,
        window_size: int,
        H: int,
        W: int,
    ) -> torch.Tensor:
        # Pure PyTorch
        if self.is_available is False or q_norm.is_cuda is False:
            return swattention_av(q_norm, attn_local, v_local, learnable_tokens, learnable_bias)

        # Custom kernel
        attn_local = (q_norm @ learnable_tokens) + learnable_bias + attn_local
        return SWAttentionFunction_AV.apply(attn_local.type_as(v_local), v_local.contiguous(), H, W, window_size)


# pylint: disable=abstract-method,arguments-differ,invalid-name
class SWAttentionFunction_QK_RPB(Function):
    @staticmethod
    def forward(ctx, query, key, rpb, height, width, kernel_size):  # type: ignore
        attn_weight = SWATTENTION.qk_rpb_forward(  # type: ignore[union-attr]
            query, key, rpb, height, width, kernel_size, SWATTENTION_CUDA_NUM_THREADS
        )

        ctx.save_for_backward(query, key)
        ctx.height = height
        ctx.width = width
        ctx.kernel_size = kernel_size

        return attn_weight

    @staticmethod
    def backward(ctx, d_attn_weight):  # type: ignore
        (query, key) = ctx.saved_tensors
        height = ctx.height
        width = ctx.width
        kernel_size = ctx.kernel_size

        (d_query, d_key, d_rpb) = SWATTENTION.qk_rpb_backward(  # type: ignore[union-attr]
            d_attn_weight.contiguous(), query, key, height, width, kernel_size, SWATTENTION_CUDA_NUM_THREADS
        )

        return (d_query, d_key, d_rpb, None, None, None)


# pylint: disable=abstract-method,arguments-differ,invalid-name
class SWAttentionFunction_AV(torch.autograd.Function):
    @staticmethod
    def forward(ctx, attn_weight, value, height, width, kernel_size):  # type: ignore
        output = SWATTENTION.av_forward(  # type: ignore[union-attr]
            attn_weight, value, height, width, kernel_size, SWATTENTION_CUDA_NUM_THREADS
        )

        ctx.save_for_backward(attn_weight, value)
        ctx.height = height
        ctx.width = width
        ctx.kernel_size = kernel_size

        return output

    @staticmethod
    def backward(ctx, d_output):  # type: ignore
        (attn_weight, value) = ctx.saved_tensors
        height = ctx.height
        width = ctx.width
        kernel_size = ctx.kernel_size

        d_attn_weight, d_value = SWATTENTION.av_backward(  # type: ignore[union-attr]
            d_output.contiguous(), attn_weight, value, height, width, kernel_size, SWATTENTION_CUDA_NUM_THREADS
        )

        return (d_attn_weight, d_value, None, None, None)


def swattention_qk_rpb(
    kv: torch.Tensor,
    q_norm_scaled: torch.Tensor,
    relative_pos_bias_local: torch.Tensor,
    padding_mask: torch.Tensor,
    num_heads: int,
    head_dim: int,
    window_size: int,
    local_len: int,
    H: int,
    W: int,
) -> tuple[torch.Tensor, torch.Tensor]:
    (B, N, _) = kv.size()

    # Generate unfolded keys and values and l2-normalize them
    (k_local, v_local) = kv.chunk(2, dim=-1)
    k_local = F.normalize(k_local.reshape(B, N, num_heads, head_dim), dim=-1).reshape(B, N, -1)
    kv_local = torch.concat([k_local, v_local], dim=-1).permute(0, 2, 1).reshape(B, -1, H, W)

    (k_local, v_local) = (
        F.unfold(kv_local, kernel_size=window_size, padding=window_size // 2, stride=1)
        .reshape(B, 2 * num_heads, head_dim, local_len, N)
        .permute(0, 1, 4, 2, 3)
        .chunk(2, dim=1)
    )

    # Compute local similarity
    attn_local = (
        (q_norm_scaled.unsqueeze(-2) @ k_local).squeeze(-2) + relative_pos_bias_local.unsqueeze(1)
    ).masked_fill(padding_mask, float("-inf"))

    return (attn_local, v_local)


def swattention_av(
    q_norm: torch.Tensor,
    attn_local: torch.Tensor,
    v_local: torch.Tensor,
    learnable_tokens: torch.Tensor,
    learnable_bias: torch.Tensor,
) -> torch.Tensor:
    return (
        ((q_norm @ learnable_tokens) + learnable_bias + attn_local).unsqueeze(-2) @ v_local.transpose(-2, -1)
    ).squeeze(-2)



import argparse
import logging
from typing import Any
from typing import Optional

import torch

from birder.common import cli
from birder.common import fs_ops
from birder.common import lib
from birder.common.lib import get_network_name
from birder.model_registry import registry
from birder.net.base import SignatureType

logger = logging.getLogger(__name__)


def avg_models(network: str, tag: Optional[str], reparameterized: bool, epochs: list[int], force: bool) -> None:
    device = torch.device("cpu")
    state_list = []
    aux_data = {}
    for idx, epoch in enumerate(epochs):
        network_name = get_network_name(network, tag)
        path = fs_ops.model_path(network_name, epoch=epoch)
        logger.info(f"Loading model from {path}...")

        model_dict: dict[str, Any] = torch.load(path, map_location=device, weights_only=True)
        state_list.append(model_dict["state"])

        if idx == 0:
            logger.info(f"Copying signature from epoch {epoch}")
            for key in model_dict:
                if key in ("state", "signature"):
                    continue

                aux_data[key] = model_dict[key]
                logger.info(f"Copying {key} from epoch {epoch}")

            signature: SignatureType = model_dict["signature"]
            input_channels = lib.get_channels_from_signature(signature)
            num_classes = lib.get_num_labels_from_signature(signature)
            size = lib.get_size_from_signature(signature)

            net = registry.net_factory(network, input_channels, num_classes, size=size)
            if reparameterized is True:
                net.reparameterize_model()

            net.to(device)

    # Average state
    logger.info("Calculating averages...")
    avg_state = {}
    for state_name in state_list[0]:
        params = torch.empty((len(state_list),) + state_list[0][state_name].size())

        for idx, state in enumerate(state_list):
            params[idx] = state[state_name]

        avg_state[state_name] = params.mean(axis=0)

    net.load_state_dict(avg_state)

    # Save model
    model_path = fs_ops.model_path(network_name, epoch=0)
    if model_path.exists() is True and force is False:
        logger.warning("Averaged model already exists... aborting")
        raise SystemExit(1)

    logger.info(f"Saving model checkpoint {model_path}...")
    torch.save(
        {
            "state": net.state_dict(),
            "signature": signature,
            **aux_data,
        },
        model_path,
    )


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "avg-model",
        allow_abbrev=False,
        help="create weight average model from multiple trained models",
        description="create weight average model from multiple trained models",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools avg-model --network efficientnet_v2_m --epochs 290 295 300\n"
            "python -m birder.tools avg-model --network shufflenet_v2_2_0 --epochs 95 100 100\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "-n", "--network", type=str, required=True, help="the neural network to use (i.e. resnet_v2)"
    )
    subparser.add_argument("--epochs", type=int, nargs="+", metavar="N", help="epochs to average")
    subparser.add_argument("-t", "--tag", type=str, help="model tag (from the training phase)")
    subparser.add_argument(
        "-r", "--reparameterized", default=False, action="store_true", help="load reparameterized model"
    )
    subparser.add_argument("--force", action="store_true", help="override existing model")
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    avg_models(args.network, args.tag, args.reparameterized, args.epochs, args.force)



"""
FasterNet, adapted from
https://github.com/JierunChen/FasterNet/blob/master/models/fasternet.py

Paper "Run, Don't Walk: Chasing Higher FLOPS for Faster Neural Networks",
https://arxiv.org/abs/2303.03667

Changes from original:
* No extra norm's for detection
"""

# Reference license: MIT

from collections import OrderedDict
from collections.abc import Callable
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import StochasticDepth

from birder.layers.activations import get_activation_module
from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class PartialConv(nn.Module):
    def __init__(self, dim: int, n_div: int) -> None:
        super().__init__()
        self.dim_conv3 = dim // n_div
        self.dim_untouched = dim - self.dim_conv3
        self.partial_conv3 = nn.Conv2d(
            self.dim_conv3, self.dim_conv3, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (x1, x2) = torch.split(x, [self.dim_conv3, self.dim_untouched], dim=1)
        x1 = self.partial_conv3(x1)
        x = torch.concat((x1, x2), dim=1)

        return x


class MLPBlock(nn.Module):
    def __init__(
        self, dim: int, n_div: int, mlp_ratio: float, drop_path: float, act_layer: Callable[..., nn.Module]
    ) -> None:
        super().__init__()
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.spatial_mixing = PartialConv(dim, n_div)
        self.mlp = nn.Sequential(
            nn.Conv2d(dim, mlp_hidden_dim, 1, bias=False),
            nn.BatchNorm2d(mlp_hidden_dim),
            act_layer(),
            nn.Conv2d(mlp_hidden_dim, dim, 1, bias=False),
        )
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shortcut = x
        x = self.spatial_mixing(x)
        x = shortcut + self.drop_path(self.mlp(x))
        return x


class PatchEmbed(nn.Module):
    def __init__(
        self, in_channels: int, embed_dim: int, patch_size: tuple[int, int], patch_stride: tuple[int, int]
    ) -> None:
        super().__init__()
        self.proj = nn.Conv2d(
            in_channels, embed_dim, kernel_size=patch_size, stride=patch_stride, padding=(0, 0), bias=False
        )
        self.norm = nn.BatchNorm2d(embed_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.proj(x)
        x = self.norm(x)

        return x


class PatchMerging(nn.Module):
    def __init__(self, dim: int, patch_size: tuple[int, int], patch_stride: tuple[int, int]) -> None:
        super().__init__()
        self.reduction = nn.Conv2d(
            dim, 2 * dim, kernel_size=patch_size, stride=patch_stride, padding=(0, 0), bias=False
        )
        self.norm = nn.BatchNorm2d(2 * dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.reduction(x)
        x = self.norm(x)

        return x


class FasterNetBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        depth: int,
        n_div: int,
        mlp_ratio: float,
        drop_path: list[int],
        act_layer: Callable[..., nn.Module],
    ) -> None:
        super().__init__()
        layers = []
        for i in range(depth):
            layers.append(
                MLPBlock(dim=dim, n_div=n_div, mlp_ratio=mlp_ratio, drop_path=drop_path[i], act_layer=act_layer)
            )

        self.blocks = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.blocks(x)
        return x


class FasterNet(DetectorBackbone):
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        patch_size = (4, 4)
        patch_stride = (4, 4)
        n_div = 4
        mlp_ratio = 2.0
        feature_dim = 1280
        depths: list[int] = self.config["depths"]
        embed_dim: int = self.config["embed_dim"]
        act_layer_name: str = self.config["act_layer_name"]
        drop_path_rate: float = self.config["drop_path_rate"]

        act_layer = get_activation_module(act_layer_name)

        self.stem = PatchEmbed(
            in_channels=self.input_channels, embed_dim=embed_dim, patch_size=patch_size, patch_stride=patch_stride
        )

        # Stochastic depth decay rule
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]

        num_stages = len(depths)
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        layers = []
        for i in range(num_stages):
            layers.append(
                FasterNetBlock(
                    dim=int(embed_dim * 2**i),
                    depth=depths[i],
                    n_div=n_div,
                    mlp_ratio=mlp_ratio,
                    drop_path=dpr[sum(depths[:i]) : sum(depths[: i + 1])],
                    act_layer=act_layer,
                )
            )
            stages[f"stage{i+1}"] = nn.Sequential(*layers)
            return_channels.append(int(embed_dim * 2**i))
            layers = []
            if i < num_stages - 1:
                layers.append(PatchMerging(dim=int(embed_dim * 2**i), patch_size=(2, 2), patch_stride=(2, 2)))

        num_features = int(embed_dim * 2 ** (num_stages - 1))
        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Conv2d(num_features, feature_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            act_layer(),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = feature_dim
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "fasternet_t0",
    FasterNet,
    config={"depths": [1, 2, 8, 2], "embed_dim": 40, "act_layer_name": "gelu", "drop_path_rate": 0.0},
)
registry.register_model_config(
    "fasternet_t1",
    FasterNet,
    config={"depths": [1, 2, 8, 2], "embed_dim": 64, "act_layer_name": "gelu", "drop_path_rate": 0.02},
)
registry.register_model_config(
    "fasternet_t2",
    FasterNet,
    config={"depths": [1, 2, 8, 2], "embed_dim": 96, "act_layer_name": "relu", "drop_path_rate": 0.05},
)
registry.register_model_config(
    "fasternet_s",
    FasterNet,
    config={"depths": [1, 2, 13, 2], "embed_dim": 128, "act_layer_name": "relu", "drop_path_rate": 0.1},
)
registry.register_model_config(
    "fasternet_m",
    FasterNet,
    config={"depths": [3, 4, 18, 3], "embed_dim": 144, "act_layer_name": "relu", "drop_path_rate": 0.2},
)
registry.register_model_config(
    "fasternet_l",
    FasterNet,
    config={"depths": [3, 4, 18, 3], "embed_dim": 192, "act_layer_name": "relu", "drop_path_rate": 0.3},
)

registry.register_weights(
    "fasternet_t0_il-common",
    {
        "description": "FasterNet T0 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 11.9,
                "sha256": "958bcddf6d5d272490972e7a5a1f77f0d39d65a3a9ffdb59c4b83c2d3562a2a9",
            }
        },
        "net": {"network": "fasternet_t0", "tag": "il-common"},
    },
)
registry.register_weights(
    "fasternet_t1_il-common",
    {
        "description": "FasterNet T1 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 26.0,
                "sha256": "6d7bce44e9eb02362290e9c46a80ac324176146727c16c753d2dcd0e039d155b",
            }
        },
        "net": {"network": "fasternet_t1", "tag": "il-common"},
    },
)



"""
Revitalized DenseNet, adapted from
https://github.com/naver-ai/rdnet/blob/main/rdnet/rdnet.py

Paper "DenseNets Reloaded: Paradigm Shift Beyond ResNets and ViTs", https://arxiv.org/abs/2403.19588
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from collections.abc import Callable
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torch import nn
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import LayerNorm2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType


class EffectiveSEModule(nn.Module):
    """
    From "CenterMask: Real-Time Anchor-Free Instance Segmentation" - https://arxiv.org/abs/1911.06667
    """

    def __init__(self, channels: int, activation: Callable[..., nn.Module] = nn.Hardsigmoid) -> None:
        super().__init__()
        self.fc = nn.Conv2d(channels, channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.gate = activation()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_se = x.mean((2, 3), keepdim=True)
        x_se = self.fc(x_se)
        x_se = self.gate(x_se)

        return x * x_se


class Block(nn.Sequential):
    def __init__(self, in_channels: int, inter_channels: int, out_channels: int) -> None:
        super().__init__()
        self.append(
            nn.Conv2d(in_channels, in_channels, groups=in_channels, kernel_size=(7, 7), stride=(1, 1), padding=(3, 3))
        )
        self.append(LayerNorm2d(in_channels, eps=1e-6))
        self.append(nn.Conv2d(in_channels, inter_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)))
        self.append(nn.GELU())
        self.append(nn.Conv2d(inter_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)))


class BlockESE(nn.Sequential):
    def __init__(self, in_channels: int, inter_channels: int, out_channels: int) -> None:
        super().__init__()
        self.append(
            nn.Conv2d(in_channels, in_channels, groups=in_channels, kernel_size=(7, 7), stride=(1, 1), padding=(3, 3))
        )
        self.append(LayerNorm2d(in_channels, eps=1e-6))
        self.append(nn.Conv2d(in_channels, inter_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)))
        self.append(nn.GELU())
        self.append(nn.Conv2d(inter_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)))
        self.append(EffectiveSEModule(out_channels))


class DenseBlock(nn.Module):
    def __init__(
        self,
        num_input_features: int,
        growth_rate: int,
        bottleneck_width_ratio: float,
        drop_path_rate: float,
        block_type: type[Block | BlockESE],
        ls_init_value: float,
    ) -> None:
        super().__init__()
        self.gamma = nn.Parameter(ls_init_value * torch.ones(growth_rate))
        growth_rate = int(growth_rate)
        inter_channels = int(num_input_features * bottleneck_width_ratio / 8) * 8

        self.drop_path = StochasticDepth(drop_path_rate, mode="row")

        self.layers = block_type(
            in_channels=num_input_features, inter_channels=inter_channels, out_channels=growth_rate
        )

    def forward(self, xl: list[torch.Tensor]) -> torch.Tensor:
        x = torch.concat(xl, dim=1)
        x = self.layers(x)
        x = x.mul(self.gamma.reshape(1, -1, 1, 1))
        x = self.drop_path(x)

        return x


class DenseStage(nn.Module):
    def __init__(
        self,
        num_block: int,
        num_input_features: int,
        drop_path_rates: list[int],
        growth_rate: int,
        bottleneck_width_ratio: float,
        block_name: str,
        ls_init_value: float,
    ) -> None:
        super().__init__()
        if block_name == "Block":
            block_type = Block
        elif block_name == "BlockESE":
            block_type = BlockESE
        else:
            raise ValueError(f"Unknown block_name '{block_name}'")

        self.layers = nn.ModuleList()
        for i in range(num_block):
            self.layers.append(
                DenseBlock(
                    num_input_features=num_input_features,
                    growth_rate=growth_rate,
                    bottleneck_width_ratio=bottleneck_width_ratio,
                    drop_path_rate=drop_path_rates[i],
                    block_type=block_type,
                    ls_init_value=ls_init_value,
                )
            )
            num_input_features += growth_rate

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        features = [x]
        for module in self.layers:
            new_feature = module(features)
            features.append(new_feature)

        return torch.concat(features, dim=1)


class RDNet(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.\d+\.layers\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        bottleneck_width_ratio = 4.0
        ls_init_value = 1e-6
        num_init_features: int = self.config["num_init_features"]
        growth_rates = self.config["growth_rates"]
        num_blocks_list: list[int] = self.config["num_blocks_list"]
        is_downsample_block: list[bool] = self.config["is_downsample_block"]
        transition_compression_ratio: float = self.config["transition_compression_ratio"]
        block_type_name: list[str] = self.config["block_type_name"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stem = nn.Sequential(
            nn.Conv2d(self.input_channels, num_init_features, kernel_size=(4, 4), stride=(4, 4), padding=(0, 0)),
            LayerNorm2d(num_init_features),
        )

        num_stages = len(growth_rates)
        num_features = num_init_features
        dp_rates = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(num_blocks_list)).split(num_blocks_list)]

        # Add dense blocks
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        dense_stage_layers: list[nn.Module] = []
        idx = 0
        for i in range(num_stages):
            if i > 0:
                compressed_num_features = int(num_features * transition_compression_ratio / 8) * 8
                k_size = (1, 1)
                if is_downsample_block[i] is True:
                    stages[f"stage{idx+1}"] = nn.Sequential(*dense_stage_layers)
                    return_channels.append(num_features)
                    dense_stage_layers = []
                    idx += 1
                    k_size = (2, 2)

                dense_stage_layers.append(LayerNorm2d(num_features))
                dense_stage_layers.append(
                    nn.Conv2d(num_features, compressed_num_features, kernel_size=k_size, stride=k_size, padding=(0, 0))
                )
                num_features = compressed_num_features

            dense_stage_layers.append(
                DenseStage(
                    num_block=num_blocks_list[i],
                    num_input_features=num_features,
                    drop_path_rates=dp_rates[i],
                    growth_rate=growth_rates[i],
                    bottleneck_width_ratio=bottleneck_width_ratio,
                    block_name=block_type_name[i],
                    ls_init_value=ls_init_value,
                )
            )
            num_features += num_blocks_list[i] * growth_rates[i]

        stages[f"stage{idx+1}"] = nn.Sequential(*dense_stage_layers)
        return_channels.append(num_features)

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            LayerNorm2d(num_features, eps=1e-6),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = num_features
        self.classifier = self.create_classifier()

        self.stem_stride = 4
        self.stem_width = num_init_features
        self.encoding_size = num_features

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)

            elif isinstance(m, nn.Linear):
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "rdnet_t",
    RDNet,
    config={
        # n_layer = 7
        "num_init_features": 64,
        "growth_rates": [64] + [104] + [128] * 4 + [224],
        "num_blocks_list": [3] * 7,
        "is_downsample_block": [False, True, True, False, False, False, True],
        "transition_compression_ratio": 0.5,
        "block_type_name": ["Block"] + ["Block"] + ["BlockESE"] * 4 + ["BlockESE"],
        "drop_path_rate": 0.15,
    },
)
registry.register_model_config(
    "rdnet_s",
    RDNet,
    config={
        # n_layer = 11
        "num_init_features": 72,
        "growth_rates": [64] + [128] + [128] * (11 - 4) + [240] * 2,
        "num_blocks_list": [3] * 11,
        "is_downsample_block": [False, True, True, False, False, False, False, False, False, True, False],
        "transition_compression_ratio": 0.5,
        "block_type_name": ["Block"] + ["Block"] + ["BlockESE"] * (11 - 4) + ["BlockESE"] * 2,
        "drop_path_rate": 0.35,
    },
)
registry.register_model_config(
    "rdnet_b",
    RDNet,
    config={
        # n_layer = 11
        "num_init_features": 120,
        "growth_rates": [96] + [128] + [168] * (11 - 4) + [336] * 2,
        "num_blocks_list": [3] * 11,
        "is_downsample_block": [False, True, True, False, False, False, False, False, False, True, False],
        "transition_compression_ratio": 0.5,
        "block_type_name": ["Block"] + ["Block"] + ["BlockESE"] * (11 - 4) + ["BlockESE"] * 2,
        "drop_path_rate": 0.4,
    },
)
registry.register_model_config(
    "rdnet_l",
    RDNet,
    config={
        # n_layer = 12
        "num_init_features": 144,
        "growth_rates": [128] + [192] + [256] * (12 - 4) + [360] * 2,
        "num_blocks_list": [3] * 12,
        "is_downsample_block": [False, True, True, False, False, False, False, False, False, False, True, False],
        "transition_compression_ratio": 0.5,
        "block_type_name": ["Block"] + ["Block"] + ["BlockESE"] * (12 - 4) + ["BlockESE"] * 2,
        "drop_path_rate": 0.5,
    },
)

registry.register_weights(
    "rdnet_s_arabian-peninsula256px",
    {
        "url": "https://huggingface.co/birder-project/rdnet_s_arabian-peninsula/resolve/main",
        "description": "RDNet small model trained on the arabian-peninsula dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 191.3,
                "sha256": "db19d6a637b3f88c397fb89cc97d0db99f6921c3042cf0ee0ecc98ed77012b47",
            }
        },
        "net": {"network": "rdnet_s", "tag": "arabian-peninsula256px"},
    },
)
registry.register_weights(
    "rdnet_s_arabian-peninsula",
    {
        "url": "https://huggingface.co/birder-project/rdnet_s_arabian-peninsula/resolve/main",
        "description": "RDNet small model trained on the arabian-peninsula dataset",
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 191.3,
                "sha256": "b3905cbae41dd71451d3508be45f6f2ab939bc23c49df512e5eb35513d8d6601",
            }
        },
        "net": {"network": "rdnet_s", "tag": "arabian-peninsula"},
    },
)



"""
ConvMixer, adapted from
https://github.com/locuslab/convmixer

Paper "Patches Are All You Need?", https://arxiv.org/abs/2201.09792

Changes from original:
* 768/32 uses GELU instead of ReLU
"""

# Reference license: MIT

from collections.abc import Callable
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import BaseNet


class Residual(nn.Module):
    def __init__(self, fn: Callable[[torch.Tensor], torch.Tensor]) -> None:
        super().__init__()
        self.fn = fn

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.fn(x) + x


class ConvMixer(BaseNet):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        dim: int = self.config["dim"]
        depth: int = self.config["depth"]
        kernel_size: tuple[int, int] = self.config["kernel_size"]
        patch_size: tuple[int, int] = self.config["patch_size"]

        self.stem = Conv2dNormActivation(
            self.input_channels,
            dim,
            kernel_size=patch_size,
            stride=patch_size,
            padding=(0, 0),
            activation_layer=nn.GELU,
            inplace=None,
        )

        padding = (kernel_size[0] // 2, kernel_size[1] // 2)
        self.body = nn.Sequential(
            *[
                nn.Sequential(
                    Residual(
                        Conv2dNormActivation(
                            dim,
                            dim,
                            kernel_size=kernel_size,
                            stride=(1, 1),
                            padding=padding,
                            groups=dim,
                            activation_layer=nn.GELU,
                            inplace=None,
                        )
                    ),
                    Conv2dNormActivation(
                        dim,
                        dim,
                        kernel_size=(1, 1),
                        stride=(1, 1),
                        padding=(0, 0),
                        activation_layer=nn.GELU,
                        inplace=None,
                    ),
                )
                for _ in range(depth)
            ]
        )

        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.embedding_size = dim
        self.classifier = self.create_classifier()

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "convmixer_768_32", ConvMixer, config={"dim": 768, "depth": 32, "kernel_size": (7, 7), "patch_size": (7, 7)}
)
registry.register_model_config(
    "convmixer_1024_20", ConvMixer, config={"dim": 1024, "depth": 20, "kernel_size": (9, 9), "patch_size": (14, 14)}
)
registry.register_model_config(
    "convmixer_1536_20", ConvMixer, config={"dim": 1536, "depth": 20, "kernel_size": (9, 9), "patch_size": (7, 7)}
)



import torch
import torch.nn.functional as F
from torch import nn
from torch.autograd import Function
from torch.autograd.function import once_differentiable

from birder.kernels.load_kernel import load_msda

MSDA = None


class MultiScaleDeformableAttention(nn.Module):
    """
    Deformable DETR: Deformable Transformers for End-to-End Object Detection: https://arxiv.org/abs/2010.04159

    Lazy-loading MSDA operator.

    The custom kernel is loaded on first instantiation, not at import time.
    Falls back to pure PyTorch implementation if kernel loading fails.
    """

    def __init__(self) -> None:
        super().__init__()

        global MSDA  # pylint: disable=global-statement
        if MSDA is None and not torch.jit.is_tracing() and not torch.jit.is_scripting():
            MSDA = load_msda()

        self.is_available = MSDA is not None

    def forward(
        self,
        value: torch.Tensor,
        value_spatial_shapes: torch.Tensor,
        value_level_start_index: torch.Tensor,
        sampling_locations: torch.Tensor,
        attention_weights: torch.Tensor,
        im2col_step: int,
    ) -> torch.Tensor:
        # Pure PyTorch
        if self.is_available is False or value.is_cuda is False:
            return multi_scale_deformable_attention(
                value, value_spatial_shapes, value_level_start_index, sampling_locations, attention_weights, im2col_step
            )

        # Custom kernel
        return MSDAFunction.apply(
            value, value_spatial_shapes, value_level_start_index, sampling_locations, attention_weights, im2col_step
        )


# pylint: disable=abstract-method,arguments-differ
class MSDAFunction(Function):
    @staticmethod
    @torch.compiler.disable()
    @torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float32)
    def forward(  # type: ignore
        ctx, value, value_spatial_shapes, value_level_start_index, sampling_locations, attention_weights, im2col_step
    ):
        ctx.im2col_step = im2col_step
        output = MSDA.ms_deform_attn_forward(  # type: ignore[union-attr]
            value, value_spatial_shapes, value_level_start_index, sampling_locations, attention_weights, ctx.im2col_step
        )
        ctx.save_for_backward(
            value, value_spatial_shapes, value_level_start_index, sampling_locations, attention_weights
        )
        return output

    @staticmethod
    @torch.compiler.disable()
    @torch.amp.custom_bwd(device_type="cuda")
    @once_differentiable
    def backward(ctx, grad_output):  # type: ignore
        (value, value_spatial_shapes, value_level_start_index, sampling_locations, attention_weights) = (
            ctx.saved_tensors
        )
        grad_value, grad_sampling_loc, grad_attn_weight = MSDA.ms_deform_attn_backward(  # type: ignore[union-attr]
            value,
            value_spatial_shapes,
            value_level_start_index,
            sampling_locations,
            attention_weights,
            grad_output,
            ctx.im2col_step,
        )

        return (grad_value, None, None, grad_sampling_loc, grad_attn_weight, None)


def multi_scale_deformable_attention(
    value: torch.Tensor,
    value_spatial_shapes: torch.Tensor,
    value_level_start_index: torch.Tensor,  # pylint:disable=unused-argument
    sampling_locations: torch.Tensor,
    attention_weights: torch.Tensor,
    im2col_step: int,  # pylint:disable=unused-argument
) -> torch.Tensor:
    (batch_size, _, num_heads, hidden_dim) = value.size()
    (_, num_queries, num_heads, num_levels, num_points, _) = sampling_locations.size()
    areas: list[int] = value_spatial_shapes.prod(dim=1).tolist()
    value_list = value.split(areas, dim=1)
    sampling_grids = 2 * sampling_locations - 1
    sampling_value_list = []
    for level_id, spatial_shape in enumerate(value_spatial_shapes):
        # (batch_size, height*width, num_heads, hidden_dim)
        # -> (batch_size, height*width, num_heads*hidden_dim)
        # -> (batch_size, num_heads*hidden_dim, height*width)
        # -> (batch_size*num_heads, hidden_dim, height, width)
        height = spatial_shape[0]
        width = spatial_shape[1]
        value_l_ = (
            value_list[level_id].flatten(2).transpose(1, 2).reshape(batch_size * num_heads, hidden_dim, height, width)
        )

        # (batch_size, num_queries, num_heads, num_points, 2)
        # -> (batch_size, num_heads, num_queries, num_points, 2)
        # -> (batch_size*num_heads, num_queries, num_points, 2)
        sampling_grid_l_ = sampling_grids[:, :, :, level_id].transpose(1, 2).flatten(0, 1)

        # (batch_size*num_heads, hidden_dim, num_queries, num_points)
        sampling_value_l_ = F.grid_sample(
            value_l_, sampling_grid_l_, mode="bilinear", padding_mode="zeros", align_corners=False
        )
        sampling_value_list.append(sampling_value_l_)

    # (batch_size, num_queries, num_heads, num_levels, num_points)
    # -> (batch_size, num_heads, num_queries, num_levels, num_points)
    # -> (batch_size, num_heads, 1, num_queries, num_levels*num_points)
    attention_weights = attention_weights.transpose(1, 2).reshape(
        batch_size * num_heads, 1, num_queries, num_levels * num_points
    )
    output = (
        (torch.stack(sampling_value_list, dim=-2).flatten(-2) * attention_weights)
        .sum(-1)
        .view(batch_size, num_heads * hidden_dim, num_queries)
    )

    return output.transpose(1, 2).contiguous()



"""
InceptionNeXt, adapted from
https://github.com/sail-sg/inceptionnext/blob/main/models/inceptionnext.py

Paper "InceptionNeXt: When Inception Meets ConvNeXt", https://arxiv.org/abs/2303.16900
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from collections.abc import Callable
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class InceptionDWConv2d(nn.Module):
    def __init__(self, in_channels: int, square_kernel_size: int, band_kernel_size: int, branch_ratio: float) -> None:
        super().__init__()

        branch_channels = int(in_channels * branch_ratio)
        self.dwconv_hw = nn.Conv2d(
            branch_channels,
            branch_channels,
            kernel_size=(square_kernel_size, square_kernel_size),
            stride=(1, 1),
            padding=square_kernel_size // 2,
            groups=branch_channels,
            bias=True,
        )
        self.dwconv_w = nn.Conv2d(
            branch_channels,
            branch_channels,
            kernel_size=(1, band_kernel_size),
            stride=(1, 1),
            padding=(0, band_kernel_size // 2),
            groups=branch_channels,
            bias=True,
        )
        self.dwconv_h = nn.Conv2d(
            branch_channels,
            branch_channels,
            kernel_size=(band_kernel_size, 1),
            stride=(1, 1),
            padding=(band_kernel_size // 2, 0),
            groups=branch_channels,
            bias=True,
        )
        self.split_indexes = (
            in_channels - (3 * branch_channels),
            branch_channels,
            branch_channels,
            branch_channels,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (x_id, x_hw, x_w, x_h) = torch.split(x, self.split_indexes, dim=1)
        x_hw = self.dwconv_hw(x_hw)
        x_w = self.dwconv_w(x_w)
        x_h = self.dwconv_h(x_h)

        return torch.concat((x_id, x_hw, x_w, x_h), dim=1)


class ConvMLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int,
        out_features: int,
        act_layer: Callable[..., nn.Module] = nn.GELU,
    ) -> None:
        super().__init__()
        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True)
        self.act = act_layer()
        self.fc2 = nn.Conv2d(
            hidden_features, out_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.act(x)
        x = self.fc2(x)

        return x


class InceptionNeXtBlock(nn.Module):
    def __init__(
        self,
        channels: int,
        mlp_ratio: float,
        layer_scale: float,
        band_kernel_size: int,
        branch_ratio: float,
        stochastic_depth_prob: float,
    ) -> None:
        super().__init__()
        self.block = nn.Sequential(
            InceptionDWConv2d(
                channels, square_kernel_size=3, band_kernel_size=band_kernel_size, branch_ratio=branch_ratio
            ),
            nn.BatchNorm2d(channels),
            ConvMLP(channels, hidden_features=int(mlp_ratio * channels), out_features=channels),
        )
        self.layer_scale = nn.Parameter(torch.ones(channels, 1, 1) * layer_scale, requires_grad=True)
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        x = self.layer_scale * self.block(x)
        x = self.stochastic_depth(x) + identity

        return x


class InceptionNeXtStage(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int,
        depth: int,
        drop_path_rates: list[float],
        mlp_ratio: float,
        layer_scale: float,
        band_kernel_size: int,
        branch_ratio: float,
    ) -> None:
        super().__init__()
        if stride > 1:
            self.downsample = nn.Sequential(
                nn.BatchNorm2d(in_channels),
                nn.Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size=(stride, stride),
                    stride=(stride, stride),
                    padding=(0, 0),
                    bias=True,
                ),
            )

        else:
            self.downsample = nn.Identity()

        stage_blocks = []
        for i in range(depth):
            stage_blocks.append(
                InceptionNeXtBlock(
                    channels=out_channels,
                    mlp_ratio=mlp_ratio,
                    layer_scale=layer_scale,
                    band_kernel_size=band_kernel_size,
                    branch_ratio=branch_ratio,
                    stochastic_depth_prob=drop_path_rates[i],
                )
            )

        self.blocks = nn.Sequential(*stage_blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


# pylint: disable=invalid-name
class Inception_NeXt(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        layer_scale = 1e-6
        mlp_ratios = [4, 4, 4, 3]
        channels: list[int] = self.config["channels"]
        num_layers: list[int] = self.config["num_layers"]
        band_kernel_size: int = self.config["band_kernel_size"]
        branch_ratio: float = self.config["branch_ratio"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stem = Conv2dNormActivation(
            self.input_channels,
            channels[0],
            kernel_size=(4, 4),
            stride=(4, 4),
            padding=(0, 0),
            bias=True,
            activation_layer=None,
        )

        num_stage = len(num_layers)
        dp_rates = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(num_layers)).split(num_layers)]

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        prev_chs = channels[0]
        for i in range(num_stage):
            out_chs = channels[i]
            if i > 0:
                stride = 2
            else:
                stride = 1

            stages[f"stage{i+1}"] = InceptionNeXtStage(
                prev_chs,
                out_chs,
                stride=stride,
                depth=num_layers[i],
                drop_path_rates=dp_rates[i],
                mlp_ratio=mlp_ratios[i],
                layer_scale=layer_scale,
                band_kernel_size=band_kernel_size,
                branch_ratio=branch_ratio,
            )
            return_channels.append(out_chs)
            prev_chs = out_chs

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = channels[-1]
        self.last_mlp_ratio = mlp_ratios[-1]
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def create_classifier(self, embed_dim: Optional[int] = None) -> nn.Module:
        if self.num_classes == 0:
            return nn.Identity()

        if embed_dim is None:
            embed_dim = self.embedding_size

        return nn.Sequential(
            nn.Linear(embed_dim, self.last_mlp_ratio * embed_dim),
            nn.GELU(),
            nn.LayerNorm(self.last_mlp_ratio * embed_dim, eps=1e-6),
            nn.Linear(self.last_mlp_ratio * embed_dim, self.num_classes),
        )


registry.register_model_config(
    "inception_next_a",
    Inception_NeXt,
    config={
        "channels": [40, 80, 160, 320],
        "num_layers": [2, 2, 6, 2],
        "band_kernel_size": 9,
        "branch_ratio": 0.25,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "inception_next_t",
    Inception_NeXt,
    config={
        "channels": [96, 192, 384, 768],
        "num_layers": [3, 3, 9, 3],
        "band_kernel_size": 11,
        "branch_ratio": 0.125,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "inception_next_s",
    Inception_NeXt,
    config={
        "channels": [96, 192, 384, 768],
        "num_layers": [3, 3, 27, 3],
        "band_kernel_size": 11,
        "branch_ratio": 0.125,
        "drop_path_rate": 0.3,
    },
)
registry.register_model_config(
    "inception_next_b",
    Inception_NeXt,
    config={
        "channels": [128, 256, 512, 1024],
        "num_layers": [3, 3, 27, 3],
        "band_kernel_size": 11,
        "branch_ratio": 0.125,
        "drop_path_rate": 0.4,
    },
)
registry.register_model_config(
    "inception_next_l",
    Inception_NeXt,
    config={
        "channels": [192, 384, 768, 1536],
        "num_layers": [3, 3, 27, 3],
        "band_kernel_size": 11,
        "branch_ratio": 0.125,
        "drop_path_rate": 0.5,
    },
)

registry.register_weights(
    "inception_next_a_il-common",
    {
        "description": "InceptionNeXt atto model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 13.7,
                "sha256": "c9e451c8dc44727acedfac400cee9d14eac5ba6b4edbfe79efcfd8802453c506",
            }
        },
        "net": {"network": "inception_next_a", "tag": "il-common"},
    },
)



import torch
from torch.optim.lr_scheduler import LRScheduler


class CooldownLR(LRScheduler):
    def __init__(
        self, optimizer: torch.optim.Optimizer, total_steps: int, eps: float = 1e-8, last_epoch: int = -1
    ) -> None:
        self.total_steps = total_steps
        self.eps = eps
        self.last_last_epoch = last_epoch
        self._first_call = True
        self._base_lr_updated = False
        super().__init__(optimizer, last_epoch)

    def get_lr(self) -> list[float]:
        if self._first_call is True:
            # First call is just initialization
            self._first_call = False
        elif self._base_lr_updated is False:
            # In case this is part of a SequentialLR or similar, adjust to the current learning rate
            # instead of the learning rate defined at init
            self.base_lrs = [param_group["lr"] for param_group in self.optimizer.param_groups]
            self._base_lr_updated = True
            if self.last_epoch == self.last_last_epoch:
                # Handle the step(0) crazy-ness of SequentialLR
                self.last_epoch += 1

        self.last_last_epoch = self.last_epoch
        if self.last_epoch > self.total_steps:
            return [0.0 for _ in self.base_lrs]

        decay_factor = 1.0 - (1.0 - self.eps) * (self.last_epoch / self.total_steps)
        return [base_lr * decay_factor for base_lr in self.base_lrs]



"""
DINO v2 distillation

Paper "DINOv2: Learning Robust Visual Features without Supervision", https://arxiv.org/abs/2304.07193
"""

import argparse
import json
import logging
import math
import random
import sys
import time
from collections.abc import Callable
from pathlib import Path
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import torch
import torchinfo
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from torchvision.datasets.folder import pil_loader  # Slower but Handles external dataset quirks better
from torchvision.transforms import v2
from tqdm import tqdm

import birder
from birder.common import cli
from birder.common import fs_ops
from birder.common import training_cli
from birder.common import training_utils
from birder.common.lib import format_duration
from birder.common.lib import get_mim_network_name
from birder.common.lib import get_network_name
from birder.common.lib import set_random_seeds
from birder.common.masking import BlockMasking
from birder.conf import settings
from birder.data.dataloader.webdataset import make_wds_loader
from birder.data.datasets.directory import make_image_dataset
from birder.data.datasets.directory import tv_loader
from birder.data.datasets.fake import FakeDataWithPaths
from birder.data.datasets.webdataset import make_wds_dataset
from birder.data.datasets.webdataset import prepare_wds_args
from birder.data.datasets.webdataset import wds_args_from_info
from birder.data.transforms.classification import RGBType
from birder.data.transforms.classification import get_rgb_stats
from birder.model_registry import Task
from birder.model_registry import registry
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import get_signature
from birder.net.ssl.base import get_ssl_signature
from birder.net.ssl.dino_v2 import DINOLoss
from birder.net.ssl.dino_v2 import DINOv2Student
from birder.net.ssl.dino_v2 import DINOv2Teacher
from birder.net.ssl.dino_v2 import KoLeoLoss
from birder.net.ssl.dino_v2 import iBOTPatchLoss

logger = logging.getLogger(__name__)


class DINOv2BlockMasking(BlockMasking):
    def __call__(self, num_masking_patches: int) -> torch.Tensor:
        mask = torch.zeros(*self.get_shape())
        mask_count = 0
        while mask_count < num_masking_patches:
            max_mask_patches = num_masking_patches - mask_count
            max_mask_patches = min(max_mask_patches, self.max_num_patches)

            delta = self._mask(mask, max_mask_patches)
            if delta == 0:
                break

            mask_count += delta

        return mask


class TrainTransform:
    def __init__(
        self,
        global_transform: Callable[..., torch.Tensor],
        crop_size: tuple[int, int],
        rgv_values: RGBType,
        local_crops_number: int,
    ) -> None:
        self.global_transform = global_transform
        self.local_crops_number = local_crops_number

        # Local small crops
        mean = rgv_values["mean"]
        std = rgv_values["std"]
        self.local_transform = v2.Compose(
            [
                v2.PILToTensor(),
                v2.RandomResizedCrop(crop_size, scale=(0.1, 0.35), interpolation=v2.InterpolationMode.BICUBIC),
                v2.RandomHorizontalFlip(p=0.5),
                v2.RandomApply([v2.ColorJitter(brightness=0.25, contrast=0.15, hue=0.04)], p=0.8),
                v2.RandomApply([v2.GaussianBlur(kernel_size=(3, 3), sigma=(0.5, 1.2))], p=0.5),
                v2.ToDtype(torch.float32, scale=True),
                v2.Normalize(mean=mean, std=std),
            ]
        )

    def __call__(self, image: Any) -> dict[str, list[torch.Tensor]]:
        output = {}
        output["global_crops"] = [self.global_transform(image), self.global_transform(image)]

        local_crops = []
        for _ in range(self.local_crops_number):
            local_crops.append(self.local_transform(image))

        output["local_crops"] = local_crops

        return output


class TrainCollator:
    def __init__(
        self,
        mask_generator: Callable[[int], tuple[list[torch.Tensor], list[torch.Tensor]]],
        seq_len: int,
        mask_probability: float,
        mask_ratio_tuple: tuple[float, float],
    ) -> None:
        self.mask_generator = mask_generator
        self.seq_len = seq_len
        self.mask_probability = mask_probability
        self.mask_ratio_tuple = mask_ratio_tuple

    def __call__(self, batch: Any) -> dict[str, Any]:
        n_global_crops = len(batch[0][1]["global_crops"])
        n_local_crops = len(batch[0][1]["local_crops"])

        collated_global_crops = torch.stack([s[1]["global_crops"][i] for i in range(n_global_crops) for s in batch])
        collated_local_crops = torch.stack([s[1]["local_crops"][i] for i in range(n_local_crops) for s in batch])

        B = len(collated_global_crops)
        N = self.seq_len
        n_samples_masked = int(B * self.mask_probability)
        probs = torch.linspace(*self.mask_ratio_tuple, n_samples_masked + 1)
        upper_bound = 0
        masks_list = []
        for i in range(0, n_samples_masked):
            prob_min = probs[i]
            prob_max = probs[i + 1]
            masks_list.append(self.mask_generator(int(N * random.uniform(prob_min, prob_max))))
            upper_bound += int(N * prob_max)

        for i in range(n_samples_masked, B):
            masks_list.append(self.mask_generator(0))

        random.shuffle(masks_list)

        collated_masks = torch.stack(masks_list).flatten(1)
        mask_indices_list = collated_masks.flatten().nonzero().flatten()

        masks_weight = (
            (1 / collated_masks.sum(-1).clamp(min=1.0)).unsqueeze(-1).expand_as(collated_masks)[collated_masks.bool()]
        )

        # Keep everything intact for loss calculation (with the randomness of the fake masking)
        # Just the remove actual masking
        collated_masks = torch.zeros_like(collated_masks)

        return {
            "collated_global_crops": collated_global_crops,
            "collated_local_crops": collated_local_crops,
            "collated_masks": collated_masks,
            "mask_indices_list": mask_indices_list,
            "masks_weight": masks_weight,
            "upper_bound": upper_bound,
            "n_masked_patches": torch.full((1,), fill_value=mask_indices_list.size(0), dtype=torch.long),
        }


# pylint: disable=too-many-locals,too-many-branches,too-many-statements
def train(args: argparse.Namespace) -> None:
    #
    # Initialize
    #
    training_utils.init_distributed_mode(args)
    logger.info(f"Starting training, birder version: {birder.__version__}, pytorch version: {torch.__version__}")
    training_utils.log_git_info()

    if args.size is None:
        args.size = registry.get_default_size(args.network)

    logger.info(f"Using size={args.size}")

    if args.cpu is True:
        device = torch.device("cpu")
        device_id = 0
    else:
        device = torch.device("cuda")
        device_id = torch.cuda.current_device()

    if args.use_deterministic_algorithms is True:
        torch.backends.cudnn.benchmark = False
        torch.use_deterministic_algorithms(True)
    else:
        torch.backends.cudnn.benchmark = True

    if args.seed is not None:
        set_random_seeds(args.seed)

    if args.non_interactive is True or training_utils.is_local_primary(args) is False:
        disable_tqdm = True
    elif sys.stderr.isatty() is False:
        disable_tqdm = True
    else:
        disable_tqdm = False

    # Enable or disable the autograd anomaly detection
    torch.autograd.set_detect_anomaly(args.grad_anomaly_detection)

    batch_size: int = args.batch_size
    logger.debug(f"Effective batch size = {args.batch_size * args.grad_accum_steps * args.world_size}")

    begin_epoch = 1
    epochs = args.epochs + 1
    if args.stop_epoch is None:
        args.stop_epoch = epochs
    else:
        args.stop_epoch += 1

    #
    # Initialize network
    #
    model_dtype: torch.dtype = getattr(torch, args.model_dtype)
    sample_shape = (batch_size, args.channels, *args.size)  # B, C, H, W
    backbone_name = get_network_name(args.network, tag="dino-v2-dist")
    network_name = get_mim_network_name("dino_v2_dist", encoder=args.network, tag=args.tag)

    student_backbone = registry.net_factory(args.network, sample_shape[1], 0, config=args.model_config, size=args.size)
    student_backbone_ema = registry.net_factory(
        args.network, sample_shape[1], 0, config=args.model_config, size=args.size
    )
    student_backbone_ema.load_state_dict(student_backbone.state_dict())

    teacher_backbone = registry.net_factory(
        args.teacher,
        sample_shape[1],
        0,
        config=args.teacher_model_config,
        size=args.size,
    )
    student_backbone.set_dynamic_size()
    if args.ibot_separate_head is False:
        args.ibot_out_dim = args.dino_out_dim

    student = DINOv2Student(
        student_backbone,
        config={
            "dino_out_dim": args.dino_out_dim,
            "use_bn": False,
            "num_layers": 3,
            "hidden_dim": 2048,
            "head_bottleneck_dim": args.head_bottleneck_dim,
            "ibot_separate_head": args.ibot_separate_head,
            "ibot_out_dim": args.ibot_out_dim,
        },
    )
    teacher = DINOv2Teacher(
        teacher_backbone,
        config={
            "dino_out_dim": args.dino_out_dim,
            "use_bn": False,
            "num_layers": 3,
            "hidden_dim": 2048,
            "head_bottleneck_dim": args.head_bottleneck_dim,
            "ibot_separate_head": args.ibot_separate_head,
            "ibot_out_dim": args.ibot_out_dim,
        },
    )
    teacher.eval()

    dino_loss = DINOLoss(args.dino_out_dim, student_temp=0.1, center_momentum=0.9)
    koleo_loss = KoLeoLoss()
    ibot_patch_loss = iBOTPatchLoss(args.ibot_out_dim, student_temp=0.1, center_momentum=0.9)

    net = torch.nn.ModuleDict(
        {
            "student": student,
            "student_backbone_ema": student_backbone_ema,
            "teacher": teacher,
            "dino_loss": dino_loss,
            "koleo_loss": koleo_loss,
            "ibot_patch_loss": ibot_patch_loss,
        }
    )
    net.task = teacher.task

    if args.resume_epoch is not None:
        begin_epoch = args.resume_epoch + 1
        (net, training_states) = fs_ops.load_simple_checkpoint(
            device, net, network_name, epoch=args.resume_epoch, strict=not args.non_strict_weights
        )
        student = net["student"]
        student_backbone_ema = net["student_backbone_ema"]
        teacher = net["teacher"]
        dino_loss = net["dino_loss"]
        koleo_loss = net["koleo_loss"]
        ibot_patch_loss = net["ibot_patch_loss"]

    else:
        state = fs_ops.load_state_dict(
            device,
            get_mim_network_name("dino_v2", encoder=args.teacher, tag=args.teacher_tag),
            epoch=args.teacher_epoch,
        )
        filtered_state = {}
        for name, p in state.items():
            if name.startswith("teacher.") is True:
                filtered_state[name.removeprefix("teacher.")] = p

        teacher.load_state_dict(filtered_state)
        training_states = fs_ops.TrainingStates.empty()

    assert isinstance(student_backbone, MaskedTokenRetentionMixin)
    assert isinstance(net, torch.nn.Module)

    net.to(device, dtype=model_dtype)
    if args.freeze_bn is True:
        student = training_utils.freeze_batchnorm2d(student)
        teacher = training_utils.freeze_batchnorm2d(teacher)
    elif args.sync_bn is True and args.distributed is True:
        student = torch.nn.SyncBatchNorm.convert_sync_batchnorm(student)
        teacher = torch.nn.SyncBatchNorm.convert_sync_batchnorm(teacher)

    if args.fast_matmul is True or args.amp is True:
        torch.set_float32_matmul_precision("high")

    # Compile networks
    teacher_compile_flag = args.compile is True or args.compile_teacher is True
    if args.compile is True:
        student = torch.compile(student)
        teacher = torch.compile(teacher)
    elif args.compile_teacher is True:
        teacher = torch.compile(teacher)

    #
    # Data
    #
    rgb_stats = get_rgb_stats(args.rgb_mode, args.rgb_mean, args.rgb_std)
    logger.debug(f"Using RGB stats: {rgb_stats}")

    mask_size = (args.size[0] // student_backbone.max_stride, args.size[1] // student_backbone.max_stride)
    seq_len = mask_size[0] * mask_size[1]

    mask_generator = DINOv2BlockMasking(
        mask_size, min_num_patches=4, max_num_patches=mask_size[0] * mask_size[1] // 2, min_aspect=0.33, max_aspect=3.33
    )
    training_transform = TrainTransform(
        training_utils.get_training_transform(args), args.local_crop_size, rgb_stats, args.local_crops_number
    )
    collator = TrainCollator(
        mask_generator, seq_len=seq_len, mask_probability=args.ibot_mask_probability, mask_ratio_tuple=(0.1, 0.5)
    )

    n_local_crops = args.local_crops_number
    n_global_crops = 2
    ibot_loss_scale = 1.0 / n_global_crops

    if args.use_fake_data is True:
        logger.warning("Using fake data")
        training_dataset = FakeDataWithPaths(
            10000, (args.channels, *args.size), num_classes=10, transform=training_transform
        )

    elif args.wds is True:
        wds_path: str | list[str]
        if args.wds_info is not None:
            (wds_path, dataset_size) = wds_args_from_info(args.wds_info, args.wds_split)
            if args.wds_train_size is not None:
                dataset_size = args.wds_train_size
        else:
            (wds_path, dataset_size) = prepare_wds_args(args.data_path[0], args.wds_train_size, device)

        training_dataset = make_wds_dataset(
            wds_path,
            dataset_size=dataset_size,
            shuffle=True,
            samples_names=True,
            transform=training_transform,
            img_loader=args.img_loader,
            cls_key=None,
            cache_dir=args.wds_cache_dir,
        )

    else:
        training_dataset = make_image_dataset(
            args.data_path,
            {},
            transforms=training_transform,
            loader=pil_loader if args.img_loader == "pil" else tv_loader,
        )

    logger.info(f"Using device {device}:{device_id}")
    logger.info(f"Training on {len(training_dataset):,} samples")

    # Data loaders and samplers
    if args.distributed is True:
        train_sampler = torch.utils.data.distributed.DistributedSampler(training_dataset, shuffle=True)
    else:
        train_sampler = torch.utils.data.RandomSampler(training_dataset)

    if args.wds is True:
        training_loader = make_wds_loader(
            training_dataset,
            batch_size,
            num_workers=args.num_workers,
            prefetch_factor=args.prefetch_factor,
            collate_fn=collator,
            world_size=args.world_size,
            pin_memory=True,
            drop_last=args.drop_last,
            shuffle=args.wds_extra_shuffle,
        )

    else:
        training_loader = DataLoader(
            training_dataset,
            batch_size=batch_size,
            sampler=train_sampler,
            num_workers=args.num_workers,
            prefetch_factor=args.prefetch_factor,
            collate_fn=collator,
            pin_memory=True,
            drop_last=args.drop_last,
        )

    last_batch_idx = len(training_loader) - 1

    #
    # Optimizer, learning rate scheduler and training parameter groups
    #

    # Training parameter groups
    custom_keys_weight_decay = training_utils.get_wd_custom_keys(args)
    parameters = training_utils.optimizer_parameter_groups(
        net,
        args.wd,
        norm_weight_decay=args.norm_wd,
        custom_keys_weight_decay=custom_keys_weight_decay,
        layer_decay=args.layer_decay,
        layer_decay_min_scale=args.layer_decay_min_scale,
        layer_decay_no_opt_scale=args.layer_decay_no_opt_scale,
        bias_lr=args.bias_lr,
    )

    # Learning rate scaling
    lr = training_utils.scale_lr(args)
    grad_accum_steps: int = args.grad_accum_steps

    if args.lr_scheduler_update == "epoch":
        iter_update = False
        iters_per_epoch = 1
    elif args.lr_scheduler_update == "iter":
        iter_update = True
        iters_per_epoch = math.ceil(len(training_loader) / grad_accum_steps)
    else:
        raise ValueError("Unsupported lr_scheduler_update")

    # Optimizer and learning rate scheduler
    optimizer = training_utils.get_optimizer(parameters, lr, args)
    scheduler = training_utils.get_scheduler(optimizer, iters_per_epoch, args)
    if args.compile_opt is True:
        optimizer.step = torch.compile(optimizer.step, fullgraph=False)

    # Temperature and weight decay schedule
    teacher_temp_schedule = training_utils.cosine_scheduler(
        args.teacher_temp,
        args.teacher_temp,
        args.epochs,
        args.warmup_teacher_temp_epochs,
        last_batch_idx + 1,
        args.warmup_teacher_temp,
    )
    if args.wd_end is not None:
        wd_schedule = training_utils.cosine_scheduler(args.wd, args.wd_end, args.epochs, 0, 1)
    else:
        wd_schedule = None

    # Gradient scaler and AMP related tasks
    (scaler, amp_dtype) = training_utils.get_amp_scaler(args.amp, args.amp_dtype)

    # Load states
    if args.load_states is True:
        optimizer.load_state_dict(training_states.optimizer_state)
        scheduler.load_state_dict(training_states.scheduler_state)
        if scaler is not None:
            scaler.load_state_dict(training_states.scaler_state)

    elif args.load_scheduler is True:
        scheduler.load_state_dict(training_states.scheduler_state)
        last_lrs = scheduler.get_last_lr()
        for g, last_lr in zip(optimizer.param_groups, last_lrs):
            g["lr"] = last_lr

    last_lr = max(scheduler.get_last_lr())
    if args.plot_lr is True:
        logger.info("Fast forwarding scheduler...")
        optimizer.step()
        lrs = []
        for _ in range(begin_epoch, epochs):
            for _ in range(iters_per_epoch):
                lrs.append(max(scheduler.get_last_lr()))
                scheduler.step()

        plt.plot(np.linspace(begin_epoch, epochs, iters_per_epoch * (epochs - begin_epoch), endpoint=False), lrs)
        plt.show()
        raise SystemExit(0)

    #
    # Distributed (DDP)
    #

    # There is no backpropagation through the teacher
    for p in teacher.parameters():
        p.requires_grad = False

    if args.distributed is True:
        student = torch.nn.parallel.DistributedDataParallel(
            student, device_ids=[args.local_rank], find_unused_parameters=args.find_unused_parameters
        )

    model_to_save = net
    if teacher_compile_flag is True and hasattr(model_to_save["teacher"], "_orig_mod") is True:
        model_to_save["teacher"] = model_to_save["teacher"]._orig_mod  # pylint: disable=protected-access
    if args.compile is True and hasattr(model_to_save["student"], "_orig_mod") is True:
        model_to_save["student"] = model_to_save["student"]._orig_mod  # pylint: disable=protected-access

    #
    # Misc
    #

    # Print network summary
    net_for_info = teacher
    if teacher_compile_flag is True and hasattr(teacher, "_orig_mod") is True:
        net_for_info = teacher._orig_mod  # pylint: disable=protected-access

    if args.no_summary is False:
        mask_indices_list = torch.tensor([0, 1])
        upper_bound = 3
        summary = torchinfo.summary(
            net_for_info,
            device=device,
            input_data={
                "x": torch.rand(sample_shape),
                "n_crops": sample_shape[0] // 2,
                "upper_bound": upper_bound,
                "mask_indices_list": mask_indices_list,
            },
            dtypes=[model_dtype],
            col_names=["input_size", "output_size", "kernel_size", "num_params"],
            depth=4,
            verbose=0,
        )
        if training_utils.is_global_primary(args) is True:
            # Write to stderr, same as all the logs
            print(summary, file=sys.stderr)

    # Training logs
    training_log_path = training_utils.training_log_path(network_name, device, args.experiment)
    logger.info(f"Logging training run at {training_log_path}")
    summary_writer = SummaryWriter(training_log_path)

    signature = get_ssl_signature(input_shape=sample_shape)
    backbone_signature = get_signature(input_shape=sample_shape, num_outputs=0)
    file_handler: logging.Handler = logging.NullHandler()
    if training_utils.is_local_primary(args) is True:
        summary_writer.flush()
        fs_ops.write_config(network_name, net_for_info, signature=signature, rgb_stats=rgb_stats)
        file_handler = training_utils.setup_file_logging(training_log_path.joinpath("training.log"))
        with open(training_log_path.joinpath("training_args.json"), "w", encoding="utf-8") as handle:
            json.dump(
                {
                    "birder_version": birder.__version__,
                    "pytorch_version": torch.__version__,
                    "cmdline": " ".join(sys.argv),
                    **vars(args),
                },
                handle,
                indent=2,
            )

        with open(training_log_path.joinpath("training_data.json"), "w", encoding="utf-8") as handle:
            json.dump(
                {"training_samples": len(training_dataset)},
                handle,
                indent=2,
            )

    #
    # Training loop
    #
    logger.info(f"Starting training with learning rate of {last_lr}")
    for epoch in range(begin_epoch, args.stop_epoch):
        tic = time.time()
        net.train()
        running_loss = training_utils.SmoothedValue()
        running_loss_dino_local = training_utils.SmoothedValue()
        running_loss_dino_global = training_utils.SmoothedValue()
        running_loss_koleo = training_utils.SmoothedValue()
        running_loss_ibot_patch = training_utils.SmoothedValue()

        if args.distributed is True:
            train_sampler.set_epoch(epoch)

        if wd_schedule is not None:
            wd = wd_schedule[epoch - 1]
            for param_group in optimizer.param_groups:
                if param_group["weight_decay"] > 0:
                    param_group["weight_decay"] = wd

            logger.info(f"Updated wd to: {wd}")

        progress = tqdm(
            desc=f"Epoch {epoch}/{epochs-1}",
            total=len(training_dataset),
            leave=False,
            disable=disable_tqdm,
            unit="samples",
            initial=0,
        )

        # Zero the parameter gradients
        optimizer.zero_grad()

        epoch_start = time.time()
        start_time = epoch_start
        last_idx = 0
        for i, data in enumerate(training_loader):
            global_step = ((epoch - 1) * (last_batch_idx + 1)) + i
            optimizer_update = (i == last_batch_idx) or ((i + 1) % grad_accum_steps == 0)
            teacher_temp = teacher_temp_schedule[global_step]

            global_crops = data["collated_global_crops"].to(device, dtype=model_dtype, non_blocking=True)
            local_crops = data["collated_local_crops"].to(device, dtype=model_dtype, non_blocking=True)

            masks = data["collated_masks"].to(device, non_blocking=True)
            mask_indices_list = data["mask_indices_list"].to(device, non_blocking=True)
            n_masked_patches_tensor = data["n_masked_patches"].to(device, non_blocking=True)
            n_masked_patches = mask_indices_list.size(0)
            upper_bound = data["upper_bound"]
            masks_weight = data["masks_weight"].to(device, non_blocking=True)

            n_local_crops_loss_terms = max(n_local_crops * n_global_crops, 1)
            n_global_crops_loss_terms = (n_global_crops - 1) * n_global_crops

            # Forward, backward and optimize
            with torch.amp.autocast("cuda", enabled=args.amp, dtype=amp_dtype):
                # Teacher
                (teacher_embedding_after_head, teacher_masked_patch_tokens_after_head) = teacher(
                    global_crops, n_global_crops, upper_bound, mask_indices_list
                )
                if args.centering == "centering":
                    teacher_dino_softmax_centered_list = dino_loss.softmax_center_teacher(
                        teacher_embedding_after_head, teacher_temp=teacher_temp
                    ).view(n_global_crops, -1, *teacher_embedding_after_head.shape[1:])
                    dino_loss.update_center(teacher_embedding_after_head)

                    teacher_masked_patch_tokens_after_head = teacher_masked_patch_tokens_after_head.unsqueeze(0)
                    masked_teacher_ibot_softmax_centered = ibot_patch_loss.softmax_center_teacher(
                        teacher_masked_patch_tokens_after_head[:, :n_masked_patches], teacher_temp=teacher_temp
                    )
                    masked_teacher_ibot_softmax_centered = masked_teacher_ibot_softmax_centered.squeeze(0)
                    ibot_patch_loss.update_center(teacher_masked_patch_tokens_after_head[:n_masked_patches])

                else:  # sinkhorn_knopp
                    teacher_dino_softmax_centered_list = dino_loss.sinkhorn_knopp_teacher(
                        teacher_embedding_after_head, teacher_temp=teacher_temp
                    ).view(n_global_crops, -1, *teacher_embedding_after_head.shape[1:])

                    masked_teacher_ibot_softmax_centered = ibot_patch_loss.sinkhorn_knopp_teacher(
                        teacher_masked_patch_tokens_after_head,
                        teacher_temp=teacher_temp,
                        n_masked_patches_tensor=n_masked_patches_tensor,
                    )

                # Student
                (
                    student_global_embedding,
                    student_global_embedding_after_head,
                    student_local_embedding_after_head,
                    student_global_masked_patch_tokens_after_head,
                ) = student(global_crops, local_crops, masks, upper_bound, mask_indices_list)

                # Local DINO loss
                loss_dino_local_crops = dino_loss(
                    student_local_embedding_after_head.chunk(n_local_crops),
                    teacher_dino_softmax_centered_list,
                ) / (n_global_crops_loss_terms + n_local_crops_loss_terms)
                loss = args.dino_loss_weight * loss_dino_local_crops

                # Global DINO loss
                loss_scales = n_global_crops
                loss_dino_global_crops = (
                    dino_loss(
                        [student_global_embedding_after_head],
                        [
                            teacher_dino_softmax_centered_list.flatten(0, 1)
                        ],  # These were chunked and stacked in reverse so A is matched to B
                    )
                    * loss_scales
                    / (n_global_crops_loss_terms + n_local_crops_loss_terms)
                )
                loss += args.dino_loss_weight * loss_dino_global_crops

                # KoLeo loss
                loss_koleo = sum(koleo_loss(p) for p in student_global_embedding.chunk(n_global_crops))
                loss += args.koleo_loss_weight * loss_koleo

                # iBOT loss
                loss_ibot_patch = (
                    ibot_patch_loss.forward_masked(
                        student_global_masked_patch_tokens_after_head,
                        masked_teacher_ibot_softmax_centered,
                        student_masks_flat=masks,
                        masks_weight=masks_weight,
                        n_masked_patches=n_masked_patches,
                    )
                    * loss_scales
                    * ibot_loss_scale
                )
                loss += args.ibot_loss_weight * loss_ibot_patch

            if scaler is not None:
                scaler.scale(loss).backward()
                if optimizer_update is True:
                    if args.clip_grad_norm is not None:
                        scaler.unscale_(optimizer)
                        torch.nn.utils.clip_grad_norm_(net.parameters(), args.clip_grad_norm)

                    scaler.step(optimizer)
                    scaler.update()
                    optimizer.zero_grad()
                    if iter_update is True:
                        scheduler.step()

            else:
                loss.backward()
                if optimizer_update is True:
                    if args.clip_grad_norm is not None:
                        torch.nn.utils.clip_grad_norm_(net.parameters(), args.clip_grad_norm)

                    optimizer.step()
                    optimizer.zero_grad()
                    if iter_update is True:
                        scheduler.step()

            if optimizer_update is True:
                # EMA update for the student backbone
                with torch.no_grad():
                    m = 0.999
                    for param_q, param_k in zip(student.backbone.parameters(), student_backbone_ema.parameters()):
                        param_k.data.mul_(m).add_((1 - m) * param_q.detach().data)

            # Statistics
            running_loss.update(loss.detach())
            running_loss_dino_local.update(loss_dino_local_crops.detach())
            running_loss_dino_global.update(loss_dino_global_crops.detach())
            running_loss_koleo.update(loss_koleo.detach())
            running_loss_ibot_patch.update(loss_ibot_patch.detach())

            # Write statistics
            if (i == last_batch_idx) or (i + 1) % args.log_interval == 0:
                time_now = time.time()
                time_cost = time_now - start_time
                steps_processed_in_interval = i - last_idx
                rate = steps_processed_in_interval * (batch_size * args.world_size) / time_cost

                avg_time_per_step = time_cost / steps_processed_in_interval
                remaining_steps_in_epoch = last_batch_idx - i
                estimated_time_to_finish_epoch = remaining_steps_in_epoch * avg_time_per_step

                start_time = time_now
                last_idx = i
                cur_lr = max(scheduler.get_last_lr())

                running_loss.synchronize_between_processes(device)
                running_loss_dino_local.synchronize_between_processes(device)
                running_loss_dino_global.synchronize_between_processes(device)
                running_loss_koleo.synchronize_between_processes(device)
                running_loss_ibot_patch.synchronize_between_processes(device)
                with training_utils.single_handler_logging(logger, file_handler, enabled=not disable_tqdm) as log:
                    log.info(
                        f"[Trn] Epoch {epoch}/{epochs-1}, step {i+1}/{last_batch_idx+1}  "
                        f"Loss: {running_loss.avg:.4f}  "
                        f"Elapsed: {format_duration(time_now-epoch_start)}  "
                        f"ETA: {format_duration(estimated_time_to_finish_epoch)}  "
                        f"T: {time_cost:.1f}s  "
                        f"R: {rate:.1f} samples/s  "
                        f"LR: {cur_lr:.4e}"
                    )

                if training_utils.is_local_primary(args) is True:
                    summary_writer.add_scalars(
                        "loss",
                        {
                            "training": running_loss.avg,
                            "local": running_loss_dino_local.avg,
                            "global": running_loss_dino_global.avg,
                            "koleo": running_loss_koleo.avg,
                            "patch": running_loss_ibot_patch.avg,
                        },
                        ((epoch - 1) * len(training_dataset)) + (i * batch_size * args.world_size),
                    )

            # Update progress bar
            progress.update(n=batch_size * args.world_size)

        progress.close()

        # Epoch training metrics
        logger.info(f"[Trn] Epoch {epoch}/{epochs-1} training_loss: {running_loss.global_avg:.4f}")
        logger.info(f"[Trn] Epoch {epoch}/{epochs-1} dino_local_loss: {running_loss_dino_local.global_avg:.4f}")
        logger.info(f"[Trn] Epoch {epoch}/{epochs-1} dino_global_loss: {running_loss_dino_global.global_avg:.4f}")
        logger.info(f"[Trn] Epoch {epoch}/{epochs-1} koleo_loss: {running_loss_koleo.global_avg:.4f}")
        logger.info(f"[Trn] Epoch {epoch}/{epochs-1} ibot_patch_loss: {running_loss_ibot_patch.global_avg:.4f}")

        # Learning rate scheduler update
        if iter_update is False:
            scheduler.step()
        if last_lr != max(scheduler.get_last_lr()):
            last_lr = max(scheduler.get_last_lr())
            logger.info(f"Updated learning rate to: {last_lr}")

        if training_utils.is_local_primary(args) is True:
            # Checkpoint model
            if epoch % args.save_frequency == 0:
                fs_ops.checkpoint_model(
                    network_name,
                    epoch,
                    model_to_save,
                    signature,
                    {},
                    rgb_stats,
                    optimizer,
                    scheduler,
                    scaler,
                    None,
                )
                fs_ops.checkpoint_model(
                    backbone_name,
                    epoch,
                    student_backbone_ema,
                    backbone_signature,
                    {},
                    rgb_stats,
                    optimizer=None,
                    scheduler=None,
                    scaler=None,
                    model_base=None,
                )
                if args.keep_last is not None:
                    fs_ops.clean_checkpoints(network_name, args.keep_last)
                    fs_ops.clean_checkpoints(backbone_name, args.keep_last)

        # Epoch timing
        toc = time.time()
        logger.info(f"Total time: {format_duration(toc - tic)}")
        logger.info("---")

        # Reset counters
        epoch_start = time.time()
        start_time = epoch_start
        last_idx = 0

    summary_writer.close()

    # Checkpoint model
    if training_utils.is_local_primary(args) is True:
        fs_ops.checkpoint_model(
            network_name,
            epoch,
            model_to_save,
            signature,
            {},
            rgb_stats,
            optimizer,
            scheduler,
            scaler,
            None,
        )
        fs_ops.checkpoint_model(
            backbone_name,
            epoch,
            student_backbone_ema,
            backbone_signature,
            {},
            rgb_stats,
            optimizer=None,
            scheduler=None,
            scaler=None,
            model_base=None,
        )

    training_utils.shutdown_distributed_mode(args)


def get_args_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="Pre-train model",
        epilog=(
            "Usage examples\n"
            "==============\n"
            "torchrun --nproc_per_node=2 -m birder.scripts.train_dino_v2_dist \\\n"
            "    --network vit_s16 \\\n"
            "    --teacher vit_l16 \\\n"
            "    --ibot-separate-head \\\n"
            "    --local-crop-size 98 98 \\\n"
            "    --centering sinkhorn_knopp \\\n"
            "    --opt adamw \\\n"
            "    --lr 0.0002 \\\n"
            "    --lr-scheduler cosine \\\n"
            "    --lr-cosine-min 1e-6 \\\n"
            "    --epochs 100 \\\n"
            "    --warmup-epochs 10 \\\n"
            "    --batch-size 16 \\\n"
            "    --wd 0.04 \\\n"
            "    --wd-end 0.2 \\\n"
            "    --norm-wd 0 \\\n"
            "    --clip-grad-norm 3 \\\n"
            "    --amp --amp-dtype bfloat16 \\\n"
            "    --compile \\\n"
            "    --data-path data/training data/raw_data\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    parser.add_argument("-n", "--network", type=str, help="the neural network to use")
    parser.add_argument(
        "--model-config",
        action=cli.FlexibleDictAction,
        help=(
            "override the model default configuration, accepts key-value pairs or JSON "
            "('drop_path_rate=0.2' or '{\"units\": [3, 24, 36, 3], \"dropout\": 0.2}'"
        ),
    )
    parser.add_argument("-t", "--tag", type=str, help="add model tag")
    parser.add_argument("--teacher", type=str, help="the neural network to use as teacher")
    parser.add_argument(
        "--teacher-model-config",
        action=cli.FlexibleDictAction,
        help=(
            "override the teacher backbone default configuration, accepts key-value pairs or JSON "
            "('drop_path_rate=0.2' or '{\"units\": [3, 24, 36, 3], \"dropout\": 0.2}'"
        ),
    )
    parser.add_argument("--teacher-tag", type=str, help="teacher training logs tag")
    parser.add_argument("--teacher-epoch", type=int, metavar="N", help="load teacher weights from selected epoch")
    parser.add_argument("--dino-loss-weight", type=float, default=1.0, help="weight for the DINO loss component")
    parser.add_argument("--dino-out-dim", type=int, default=65536, help="dimensionality of the DINO head output")
    parser.add_argument("--head-bottleneck-dim", type=int, default=256, help="dimensionality of heads output")
    parser.add_argument("--koleo-loss-weight", type=float, default=0.1, help="weight for the KoLeo regularization loss")
    parser.add_argument("--ibot-loss-weight", type=float, default=1.0, help="weight for the iBOT loss component")
    parser.add_argument(
        "--ibot-mask-probability", type=float, default=0.5, help="probability of applying masking for iBOT training"
    )
    parser.add_argument(
        "--ibot-separate-head", default=False, action="store_true", help="use separate head for iBOT loss computation"
    )
    parser.add_argument("--ibot-out-dim", type=int, default=65536, help="dimensionality of the iBOT head output")
    parser.add_argument(
        "--warmup-teacher-temp",
        type=float,
        default=0.04,
        help="initial value for the teacher temperature, try decreasing it if the training loss does not decrease",
    )
    parser.add_argument(
        "--teacher-temp", type=float, default=0.07, help="final value (after linear warmup) of the teacher temperature"
    )
    parser.add_argument(
        "--warmup-teacher-temp-epochs", type=int, default=30, help="number of warmup epochs for the teacher temperature"
    )
    parser.add_argument(
        "--freeze-last-layer-epochs",
        default=1,
        type=int,
        help=(
            "number of epochs during which the output layer is frozen, "
            "try increasing this value if the loss does not decrease"
        ),
    )
    parser.add_argument("--local-crops-number", type=int, default=8, help="number of small local views to generate")
    parser.add_argument(
        "--local-crop-size", type=int, nargs="+", default=[96, 96], metavar=("H", "W"), help="local view size"
    )
    parser.add_argument(
        "--centering",
        type=str,
        choices=["centering", "sinkhorn_knopp"],
        default="centering",
        help="algorithm for centering",
    )
    training_cli.add_optimization_args(parser)
    training_cli.add_lr_wd_args(parser, wd_end=True)
    training_cli.add_lr_scheduler_args(parser)
    training_cli.add_training_schedule_args(parser, default_epochs=200)
    training_cli.add_input_args(parser)
    training_cli.add_data_aug_args(parser, default_level=5, default_min_scale=0.35, default_re_prob=0.0)
    training_cli.add_dataloader_args(parser, default_drop_last=True)
    training_cli.add_batch_norm_args(parser)
    training_cli.add_precision_args(parser)
    training_cli.add_compile_args(parser, teacher=True)
    training_cli.add_checkpoint_args(parser)
    training_cli.add_distributed_args(parser)
    training_cli.add_logging_and_debug_args(parser, default_log_interval=100)
    training_cli.add_training_data_args(parser, unsupervised=True)

    return parser


def validate_args(args: argparse.Namespace) -> None:
    args.data_path = [str(p) for p in args.data_path]
    args.size = cli.parse_size(args.size)
    args.local_crop_size = cli.parse_size(args.local_crop_size)

    # This will capture the common argument mistakes
    training_cli.common_args_validation(args)

    # Script specific checks
    if registry.exists(args.network, task=Task.IMAGE_CLASSIFICATION, net_type=MaskedTokenRetentionMixin) is False:
        raise cli.ValidationError(f"--network {args.network} not supported, see list-models tool for available options")
    if registry.exists(args.teacher, task=Task.IMAGE_CLASSIFICATION, net_type=MaskedTokenRetentionMixin) is False:
        raise cli.ValidationError(f"--teacher {args.teacher} not supported, see list-models tool for available options")


def args_from_dict(**kwargs: Any) -> argparse.Namespace:
    parser = get_args_parser()
    parser.set_defaults(**kwargs)
    args = parser.parse_args([])
    validate_args(args)

    return args


def main() -> None:
    parser = get_args_parser()
    args = parser.parse_args()
    validate_args(args)

    if settings.MODELS_DIR.exists() is False:
        logger.info(f"Creating {settings.MODELS_DIR} directory...")
        settings.MODELS_DIR.mkdir(parents=True, exist_ok=True)

    if args.wds_cache_dir is not None and Path(args.wds_cache_dir).exists() is False:
        logger.info(f"Creating {args.wds_cache_dir} directory...")
        Path(args.wds_cache_dir).mkdir(parents=True, exist_ok=True)

    train(args)


if __name__ == "__main__":
    logger = logging.getLogger(getattr(__spec__, "name", __name__))
    main()



"""
Paper "FlexiViT: One Model for All Patch Sizes", https://arxiv.org/abs/2212.08013
"""

# Reference license: Apache-2.0

import logging
import math
import random
from functools import partial
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from birder.common.masking import mask_tensor
from birder.layers import FFN
from birder.layers import MultiHeadAttentionPool
from birder.layers import SwiGLU_FFN
from birder.layers.activations import get_activation_module
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenOmissionMixin
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenOmissionResultType
from birder.net.base import TokenRetentionResultType
from birder.net.vit import Encoder
from birder.net.vit import EncoderBlock
from birder.net.vit import PatchEmbed
from birder.net.vit import adjust_position_embedding

logger = logging.getLogger(__name__)


def get_patch_sizes(min_size: int, max_size: int, input_size: tuple[int, int]) -> list[int]:
    (H, W) = input_size
    valid_sizes = []
    for patch_size in range(min_size, max_size + 1):
        if H % patch_size == 0 and W % patch_size == 0:
            valid_sizes.append(patch_size)

    return sorted(valid_sizes)


# No compile support for antialias
@torch.compiler.disable()  # type: ignore[misc]
def interpolate_proj(proj_weight: torch.Tensor, patch_size: int) -> torch.Tensor:
    orig_dtype = proj_weight.dtype
    proj_weight = proj_weight.float()  # Interpolate needs float32
    weight_resampled = F.interpolate(proj_weight, size=(patch_size, patch_size), mode="bicubic", antialias=True)
    weight_resampled = weight_resampled.to(orig_dtype)

    return weight_resampled


def flex_proj(
    x: torch.Tensor, proj_weight: torch.Tensor, proj_bias: Optional[torch.Tensor], patch_size: Optional[int]
) -> torch.Tensor:
    if patch_size is not None and patch_size != proj_weight.shape[-1] and not torch.jit.is_scripting():
        weight_resampled = interpolate_proj(proj_weight, patch_size)
        x = F.conv2d(x, weight_resampled, proj_bias, stride=(patch_size, patch_size))  # pylint: disable=not-callable
    else:
        x = F.conv2d(x, proj_weight, proj_bias, stride=proj_weight.shape[-2:])  # pylint: disable=not-callable

    return x


# pylint: disable=too-many-instance-attributes
class FlexiViT(DetectorBackbone, PreTrainEncoder, MaskedTokenOmissionMixin, MaskedTokenRetentionMixin):
    default_size = (240, 240)
    block_group_regex = r"encoder\.block\.(\d+)"

    # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        image_size = self.size
        attention_dropout = 0.0
        dropout = 0.0
        pos_embed_special_tokens: bool = self.config.get("pos_embed_special_tokens", False)
        patch_size: int = self.config["patch_size"]
        num_layers: int = self.config["num_layers"]
        num_heads: int = self.config["num_heads"]
        hidden_dim: int = self.config["hidden_dim"]
        mlp_dim: int = self.config["mlp_dim"]
        layer_scale_init_value: Optional[float] = self.config.get("layer_scale_init_value", None)
        pre_norm: bool = self.config.get("pre_norm", False)
        post_norm: bool = self.config.get("post_norm", True)
        num_reg_tokens: int = self.config.get("num_reg_tokens", 0)
        class_token: bool = self.config.get("class_token", True)
        attn_pool_head: bool = self.config.get("attn_pool_head", False)
        attn_pool_num_heads: Optional[int] = self.config.get("attn_pool_num_heads", None)
        attn_pool_special_tokens: bool = self.config.get("attn_pool_special_tokens", False)
        norm_layer_type: str = self.config.get("norm_layer_type", "LayerNorm")
        norm_layer_eps: float = self.config.get("norm_layer_eps", 1e-6)
        mlp_layer_type: str = self.config.get("mlp_layer_type", "FFN")
        act_layer_type: Optional[str] = self.config.get("act_layer_type", None)  # Default according to mlp type
        min_patch_size: int = self.config.get("min_patch_size", 8)
        max_patch_size: int = self.config.get("max_patch_size", 48)
        drop_path_rate: float = self.config["drop_path_rate"]

        if norm_layer_type == "LayerNorm":
            norm_layer = nn.LayerNorm
        elif norm_layer_type == "RMSNorm":
            norm_layer = nn.RMSNorm
        else:
            raise ValueError(f"Unknown norm_layer_type '{norm_layer_type}'")

        if mlp_layer_type == "FFN":
            mlp_layer = FFN
            act_layer = nn.GELU
        elif mlp_layer_type == "SwiGLU_FFN":
            mlp_layer = SwiGLU_FFN
            act_layer = nn.SiLU
        else:
            raise ValueError(f"Unknown mlp_layer_type '{mlp_layer_type}'")

        if act_layer_type is not None:
            act_layer = get_activation_module(act_layer_type)

        torch._assert(image_size[0] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(image_size[1] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(hidden_dim % num_heads == 0, "Hidden dim indivisible by num heads!")
        self.pos_embed_special_tokens = pos_embed_special_tokens
        self.patch_size = patch_size
        self.num_layers = num_layers
        self.hidden_dim = hidden_dim
        self.num_reg_tokens = num_reg_tokens
        self.attn_pool_special_tokens = attn_pool_special_tokens
        self.patch_size_list = get_patch_sizes(min_patch_size, max_patch_size, self.size)
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, num_layers)]  # Stochastic depth decay rule

        self.conv_proj = nn.Conv2d(
            self.input_channels,
            hidden_dim,
            kernel_size=(patch_size, patch_size),
            stride=(patch_size, patch_size),
            padding=(0, 0),
            bias=not pre_norm,
        )
        self.patch_embed = PatchEmbed()

        seq_length = (image_size[0] // patch_size) * (image_size[1] // patch_size)
        self.num_special_tokens = 0

        # Add a class token
        if class_token is True:
            self.class_token = nn.Parameter(torch.zeros(1, 1, hidden_dim))
            self.num_special_tokens += 1
            if pos_embed_special_tokens is True:
                seq_length += 1
        else:
            self.class_token = None

        # Add optional register tokens
        if self.num_reg_tokens > 0:
            self.reg_tokens = nn.Parameter(torch.zeros(1, self.num_reg_tokens, hidden_dim))
            self.num_special_tokens += self.num_reg_tokens
            if pos_embed_special_tokens is True:
                seq_length += self.num_reg_tokens
        else:
            self.reg_tokens = None

        # Add positional embedding
        self.pos_embedding = nn.Parameter(torch.empty(1, seq_length, hidden_dim).normal_(std=0.02))

        # Encoder
        self.encoder = Encoder(
            num_layers,
            num_heads,
            hidden_dim,
            mlp_dim,
            dropout,
            attention_dropout,
            dpr,
            pre_norm=pre_norm,
            activation_layer=act_layer,
            layer_scale_init_value=layer_scale_init_value,
            norm_layer=norm_layer,
            norm_layer_eps=norm_layer_eps,
            mlp_layer=mlp_layer,
        )

        if post_norm is True:
            self.norm = norm_layer(hidden_dim, eps=norm_layer_eps)
        else:
            self.norm = nn.Identity()

        if attn_pool_head is False:
            self.attn_pool = None
        else:
            if attn_pool_num_heads is None:
                attn_pool_num_heads = num_heads

            self.attn_pool = MultiHeadAttentionPool(hidden_dim, attn_pool_num_heads, mlp_dim, qkv_bias=True)

        self.return_stages = ["neck"]  # Actually meaningless, just for completeness
        self.return_channels = [hidden_dim]
        self.embedding_size = hidden_dim
        self.classifier = self.create_classifier()

        self.max_stride = patch_size
        self.stem_stride = patch_size
        self.stem_width = hidden_dim
        self.encoding_size = hidden_dim
        self.decoder_block = partial(
            EncoderBlock,
            16,
            mlp_dim=None,
            dropout=0,
            attention_dropout=0,
            drop_path=0,
            activation_layer=act_layer,
            norm_layer=norm_layer,
            mlp_layer=mlp_layer,
        )

        self.set_dynamic_size()

        # Weight initialization
        if isinstance(self.conv_proj, nn.Conv2d):
            # Init the patchify stem
            fan_in = self.conv_proj.in_channels * self.conv_proj.kernel_size[0] * self.conv_proj.kernel_size[1]
            nn.init.trunc_normal_(self.conv_proj.weight, std=math.sqrt(1 / fan_in))
            if self.conv_proj.bias is not None:
                nn.init.zeros_(self.conv_proj.bias)

        if isinstance(self.classifier, nn.Linear):
            nn.init.zeros_(self.classifier.weight)
            nn.init.zeros_(self.classifier.bias)

    def _get_pos_embed(self, H: int, W: int, patch_size: Optional[int] = None) -> torch.Tensor:
        if patch_size is None:
            patch_size = self.patch_size

        if H == self.size[0] and W == self.size[1] and patch_size == self.patch_size:
            return self.pos_embedding

        return adjust_position_embedding(
            self.pos_embedding,
            (self.size[0] // self.patch_size, self.size[1] // self.patch_size),
            (H // patch_size, W // patch_size),
            self.num_special_tokens if self.pos_embed_special_tokens is True else 0,
            antialias=False,
        )

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad = True

        if unfreeze_features is True:
            if self.attn_pool is not None:
                for param in self.attn_pool.parameters():
                    param.requires_grad = True

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        (H, W) = x.shape[-2:]
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        if self.pos_embed_special_tokens is False:
            x = x + self._get_pos_embed(H, W)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        if self.pos_embed_special_tokens is True:
            x = x + self._get_pos_embed(H, W)

        x = self.encoder(x)
        x = self.norm(x)

        x = x[:, self.num_special_tokens :]
        x = x.permute(0, 2, 1)
        (B, C, _) = x.size()
        x = x.reshape(B, C, H // self.patch_size, W // self.patch_size)

        return {self.return_stages[0]: x}

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.conv_proj.parameters():
            param.requires_grad = False

        self.pos_embedding.requires_grad = False

        for idx, module in enumerate(self.encoder.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    # pylint: disable=too-many-branches
    def masked_encoding_omission(
        self,
        x: torch.Tensor,
        ids_keep: Optional[torch.Tensor] = None,
        return_all_features: bool = False,
        return_keys: Literal["all", "tokens", "embedding"] = "tokens",
    ) -> TokenOmissionResultType:
        (H, W) = x.shape[-2:]

        # Reshape and permute the input tensor
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        # Add pos embedding without special tokens
        pos_embedding = self._get_pos_embed(H, W)
        if self.pos_embed_special_tokens is True:
            x = x + pos_embedding[:, self.num_special_tokens :, :]
        else:
            x = x + pos_embedding

        # Mask tokens
        if ids_keep is not None:
            x = torch.gather(x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, x.size(2)))

        # Append class and register tokens
        if self.class_token is not None:
            if self.pos_embed_special_tokens is True:
                cls_token = self.class_token + pos_embedding[:, self.num_reg_tokens : self.num_reg_tokens + 1, :]
            else:
                cls_token = self.class_token

            batch_class_token = cls_token.expand(x.shape[0], -1, -1)
            x = torch.concat((batch_class_token, x), dim=1)

        if self.reg_tokens is not None:
            if self.pos_embed_special_tokens is True:
                reg_tokens = self.reg_tokens + pos_embedding[:, 0 : self.num_reg_tokens, :]
            else:
                reg_tokens = self.reg_tokens

            batch_reg_tokens = reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        # Apply transformer
        if return_all_features is True:
            xs = self.encoder.forward_features(x)
            xs[-1] = self.norm(xs[-1])
            x = torch.stack(xs, dim=-1)
        else:
            x = self.encoder(x)
            x = self.norm(x)

        result: TokenOmissionResultType = {}
        if return_keys in ("all", "tokens"):
            result["tokens"] = x

        if return_keys in ("all", "embedding"):
            if return_all_features is True:
                x = x[..., -1]

            if self.attn_pool is not None:
                if self.attn_pool_special_tokens is False:
                    x = x[:, self.num_special_tokens :]

                x = self.attn_pool(x)
                result["embedding"] = x[:, 0]
            elif self.class_token is None:
                x = x[:, self.num_special_tokens :]
                result["embedding"] = x.mean(dim=1)
            else:
                result["embedding"] = x[:, self.num_reg_tokens]

        return result

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        (H, W) = x.shape[-2:]

        x = self.conv_proj(x)
        x = mask_tensor(x, mask, mask_token=mask_token, patch_factor=self.max_stride // self.stem_stride)

        # Reshape and permute the input tensor
        x = self.patch_embed(x)

        if self.pos_embed_special_tokens is False:
            x = x + self._get_pos_embed(H, W)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        if self.pos_embed_special_tokens is True:
            x = x + self._get_pos_embed(H, W)

        x = self.encoder(x)
        x = self.norm(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            features = x[:, self.num_special_tokens :]
            features = features.permute(0, 2, 1)
            (B, C, _) = features.size()
            features = features.reshape(B, C, H // self.patch_size, W // self.patch_size)
            result["features"] = features

        if return_keys in ("all", "embedding"):
            if self.attn_pool is not None:
                if self.attn_pool_special_tokens is False:
                    x = x[:, self.num_special_tokens :]

                x = self.attn_pool(x)
                result["embedding"] = x[:, 0]
            elif self.class_token is None:
                x = x[:, self.num_special_tokens :]
                result["embedding"] = x.mean(dim=1)
            else:
                result["embedding"] = x[:, self.num_reg_tokens]

        return result

    def forward_features(self, x: torch.Tensor, patch_size: Optional[int] = None) -> torch.Tensor:
        (H, W) = x.shape[-2:]

        # Reshape and permute the input tensor
        x = flex_proj(x, self.conv_proj.weight, self.conv_proj.bias, patch_size)
        x = self.patch_embed(x)

        if self.pos_embed_special_tokens is False:
            x = x + self._get_pos_embed(H, W, patch_size=patch_size)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        if self.pos_embed_special_tokens is True:
            x = x + self._get_pos_embed(H, W, patch_size=patch_size)

        x = self.encoder(x)
        x = self.norm(x)

        return x

    def embedding(self, x: torch.Tensor, patch_size: Optional[int] = None) -> torch.Tensor:
        x = self.forward_features(x, patch_size)

        if self.attn_pool is not None:
            if self.attn_pool_special_tokens is False:
                x = x[:, self.num_special_tokens :]

            x = self.attn_pool(x)
            return x[:, 0]

        if self.class_token is None:
            x = x[:, self.num_special_tokens :]
            return x.mean(dim=1)

        # Classifier "token" as used by standard language architectures
        return x[:, self.num_reg_tokens]

    def forward(self, x: torch.Tensor, patch_size: Optional[int] = None) -> torch.Tensor:
        if self.training is True and patch_size is None and not torch.jit.is_tracing() and not torch.jit.is_scripting():
            patch_size = random.choice(self.patch_size_list)

        x = self.embedding(x, patch_size)
        return self.classify(x)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        super().set_dynamic_size(dynamic_size)
        assert self.dynamic_size is True, "FlexiViT only support dynamic mode"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        assert new_size[0] % self.patch_size == 0, "Input shape indivisible by patch size!"
        assert new_size[1] % self.patch_size == 0, "Input shape indivisible by patch size!"

        old_size = self.size
        super().adjust_size(new_size)

        if self.pos_embed_special_tokens is True:
            num_prefix_tokens = self.num_special_tokens
        else:
            num_prefix_tokens = 0

        self.pos_embedding = nn.Parameter(
            # On rounding error see: https://github.com/facebookresearch/dino/issues/8
            adjust_position_embedding(
                self.pos_embedding,
                (old_size[0] // self.patch_size, old_size[1] // self.patch_size),
                (new_size[0] // self.patch_size, new_size[1] // self.patch_size),
                num_prefix_tokens,
            )
        )

    def adjust_patch_size(self, patch_size: int) -> None:
        if self.patch_size == patch_size:
            return

        logger.debug(f"Setting patch size to: {patch_size}")
        self.conv_proj.weight = nn.Parameter(interpolate_proj(self.conv_proj.weight, patch_size))

        # Adjust pos_embedding accordingly
        if self.pos_embed_special_tokens is True:
            num_prefix_tokens = self.num_special_tokens
        else:
            num_prefix_tokens = 0

        self.pos_embedding = nn.Parameter(
            adjust_position_embedding(
                self.pos_embedding,
                (self.size[0] // self.patch_size, self.size[1] // self.patch_size),
                (self.size[0] // patch_size, self.size[1] // patch_size),
                num_prefix_tokens,
            )
        )

        self.patch_size = patch_size

    def load_vit_weights(self, state_dict: dict[str, Any]) -> None:
        num_special_tokens = 0
        if "class_token" in state_dict:
            num_special_tokens += 1

        if "reg_tokens" in state_dict:
            num_special_tokens += state_dict["reg_tokens"].size(1)

        seq_length = (self.size[0] // self.patch_size) * (self.size[1] // self.patch_size)
        vit_pos_embed_special_tokens = state_dict["pos_embedding"].size(1) != seq_length

        # Adjust pos_embedding
        if self.pos_embed_special_tokens is False and vit_pos_embed_special_tokens is True:
            if state_dict["pos_embedding"].ndim == 2:
                state_dict["pos_embedding"] = state_dict["pos_embedding"][num_special_tokens:, :]
            else:
                state_dict["pos_embedding"] = state_dict["pos_embedding"][:, num_special_tokens:, :]

        self.load_state_dict(state_dict, strict=True)


registry.register_model_config(
    "flexivit_s16",
    FlexiViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "flexivit_s16_ls",
    FlexiViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "layer_scale_init_value": 1e-5,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "flexivit_reg1_s16",
    FlexiViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "num_reg_tokens": 1,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "flexivit_reg1_s16_rms_ls",
    FlexiViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "layer_scale_init_value": 1e-5,
        "num_reg_tokens": 1,
        "norm_layer_type": "RMSNorm",
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "flexivit_reg8_b14_ap",
    FlexiViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "num_reg_tokens": 8,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)

registry.register_weights(
    "flexivit_reg1_s16_rms_ls_dino-v2-il-all",
    {
        "url": "https://huggingface.co/birder-project/flexivit_reg1_s16_rms_ls_dino-v2-il-all/resolve/main",
        "description": (
            "FlexiViT reg1 s16 RMS norm with layer scaling model pre-trained using DINOv2 on "
            "the il-all dataset and then fine-tuned on the il-all dataset"
        ),
        "resolution": (240, 240),
        "formats": {
            "pt": {
                "file_size": 83.6,
                "sha256": "8d11fb14630f2a54632aeebd09c5a9c2b3b7de1099e09de5e91f433ed915b784",
            },
        },
        "net": {"network": "flexivit_reg1_s16_rms_ls", "tag": "dino-v2-il-all"},
    },
)



"""
data2vec, adapted from
https://github.com/arxyzan/data2vec-pytorch/blob/main/data2vec/data2vec.py
and
https://github.com/facebookresearch/fairseq/blob/main/examples/data2vec/models/data2vec_vision.py

Paper "data2vec: A General Framework for Self-supervised Learning in Speech, Vision and Language",
https://arxiv.org/abs/2202.03555

Changes from original:
* The head (final_proj in original), not applied in the vision modality - applied here
"""

# Reference license: MIT (both)

import copy
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from birder.net.base import MaskedTokenOmissionMixin
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.ssl.base import SSLBaseNet


class Data2Vec(SSLBaseNet):
    def __init__(
        self,
        backbone: PreTrainEncoder,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(backbone, config=config, size=size)
        assert self.config is not None, "must set config"
        assert isinstance(self.backbone, MaskedTokenOmissionMixin)
        assert isinstance(self.backbone, MaskedTokenRetentionMixin)

        normalize_targets: bool = self.config["normalize_targets"]
        average_top_k_layers: int = self.config["average_top_k_layers"]
        loss_beta: float = self.config["loss_beta"]

        self.normalize_targets = normalize_targets
        self.average_top_k_layers = average_top_k_layers
        self.loss_beta = loss_beta

        self.ema_backbone = copy.deepcopy(self.backbone)
        self.head = nn.Linear(self.backbone.embedding_size, self.backbone.embedding_size)

        self.mask_token = nn.Parameter(torch.zeros(1, 1, 1, self.backbone.stem_width), requires_grad=True)

        # Weights initialization
        self.ema_backbone.load_state_dict(self.backbone.state_dict())

        nn.init.trunc_normal_(self.head.weight, std=0.02)
        if self.head.bias is not None:
            nn.init.zeros_(self.head.bias)

        nn.init.trunc_normal_(self.mask_token, std=0.02)

    def forward(  # type: ignore[override]  # pylint: disable=arguments-differ
        self, src: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        x = self.backbone.masked_encoding_retention(src, mask, mask_token=self.mask_token)["features"]
        x = x.flatten(2).transpose(1, 2)

        y = self.ema_backbone.masked_encoding_omission(src, return_all_features=True)["tokens"]
        y = y[:, self.ema_backbone.num_special_tokens :]
        y = y[..., -self.average_top_k_layers :]  # Take the last k layers
        y = y.permute(3, 0, 1, 2)

        y = [F.layer_norm(t.float(), t.shape[-1:]) for t in y[:-1]] + [y[-1]]
        y = sum(y) / len(y)
        if self.normalize_targets is True:
            y = F.layer_norm(y.float(), y.shape[-1:])

        mask = ~mask.bool()
        x = x[mask]
        y = y[mask]

        x = self.head(x)
        loss = F.smooth_l1_loss(x, y, reduction="none", beta=self.loss_beta).sum(dim=-1).mean()

        return loss



from birder.common.fs_ops import load_model_with_cfg
from birder.common.fs_ops import load_pretrained_model
from birder.common.lib import get_channels_from_signature
from birder.common.lib import get_size_from_signature
from birder.data.transforms.classification import inference_preset as classification_transform
from birder.data.transforms.detection import InferenceTransform as detection_transform
from birder.inference.classification import evaluate as evaluate_classification
from birder.model_registry.model_registry import list_pretrained_models
from birder.version import __version__

__all__ = [
    "classification_transform",
    "detection_transform",
    "evaluate_classification",
    "get_channels_from_signature",
    "get_size_from_signature",
    "list_pretrained_models",
    "load_model_with_cfg",
    "load_pretrained_model",
    "__version__",
]



import argparse
import fnmatch
from typing import Any

from rich.columns import Columns
from rich.console import Console
from rich.table import Table

from birder.common import cli
from birder.model_registry import Task
from birder.model_registry import registry
from birder.model_registry.model_registry import group_sort
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenOmissionMixin
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "list-models",
        allow_abbrev=False,
        help="list available models",
        description="list available models",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools list-models\n"
            "python -m birder.tools list-models --classification\n"
            "python -m birder.tools list-models --classification --detector-backbone\n"
            "python -m birder.tools list-models --pretrain-encoder\n"
            "python -m birder.tools list-models --detection\n"
            "python -m birder.tools list-models --pretrained\n"
            "python -m birder.tools list-models --pretrained\n"
            "python -m birder.tools list-models --pretrained --detection --verbose\n"
            "python tool.py list-models --pretrained --verbose --filter '*mobile*'\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )

    task_group = subparser.add_mutually_exclusive_group(required=False)
    task_group.add_argument("--classification", default=False, action="store_true", help="list classification models")
    task_group.add_argument("--detection", default=False, action="store_true", help="list detection models")
    task_group.add_argument("--mim", default=False, action="store_true", help="list MIM models")
    task_group.add_argument("--ssl", default=False, action="store_true", help="list SSL models")

    subparser.add_argument("--pretrained", default=False, action="store_true", help="list pretrained models")

    type_group = subparser.add_argument_group()
    type_group.add_argument(
        "--detector-backbone", default=False, action="store_true", help="list detector backbone models"
    )
    type_group.add_argument(
        "--pretrain-encoder", default=False, action="store_true", help="list models that support mim pretraining"
    )
    type_group.add_argument(
        "--token-omission",
        default=False,
        action="store_true",
        help="list models supporting token omission for sparse processing",
    )
    type_group.add_argument(
        "--token-retention", default=False, action="store_true", help="list models supporting masked token retention"
    )

    subparser.add_argument("--filter", type=str, help="filter results with a fnmatch type filter)")
    subparser.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="enable verbose output with additional model details",
    )
    subparser.set_defaults(func=main)


# pylint: disable=too-many-branches
def main(args: argparse.Namespace) -> None:
    types: list[type] = []
    if args.detector_backbone is True:
        types.append(DetectorBackbone)
    if args.pretrain_encoder is True:
        types.append(PreTrainEncoder)
    if args.token_omission is True:
        types.append(MaskedTokenOmissionMixin)
    if args.token_retention is True:
        types.append(MaskedTokenRetentionMixin)

    if len(types) == 0:
        t = None
    else:
        t = tuple(types)

    # Determine the task based on the selected flags
    task = None
    if args.classification is True:
        task = Task.IMAGE_CLASSIFICATION
    elif args.detection is True:
        task = Task.OBJECT_DETECTION
    elif args.mim is True:
        task = Task.MASKED_IMAGE_MODELING
    elif args.ssl is True:
        task = Task.SELF_SUPERVISED_LEARNING

    if args.pretrained is True:
        model_list = registry.list_pretrained_models(task=task)
    else:
        model_list = registry.list_models(task=task, net_type=t)

    model_list = group_sort(model_list)
    if args.filter is not None:
        model_list = fnmatch.filter(model_list, args.filter)

    console = Console()
    if args.verbose is True:
        if args.pretrained is True:
            table = Table(show_header=True, header_style="bold dark_magenta")
            table.add_column("Model name")
            table.add_column("Format", style="dim")
            table.add_column("File size", justify="right")
            table.add_column("Resolution", justify="right")
            table.add_column("Description")
            for model_name in model_list:
                model_metadata = registry.get_pretrained_metadata(model_name)
                for format_name, format_info in model_metadata["formats"].items():
                    table.add_row(
                        model_name,
                        format_name,
                        f"{format_info['file_size']}MB",
                        "x".join(str(x) for x in model_metadata["resolution"]),
                        model_metadata["description"],
                    )

            console.print(table)

        else:
            raise NotImplementedError
            # table = Table(show_header=True, header_style="bold dark_magenta")
            # table.add_column("Model name")
            # table.add_column("Description")
            # for model_name in model_list:
            #     if model_name in registry.aliases:
            #         desc = ""

            #     else:
            #         net = registry.all_nets[model_name]
            #         desc = sys.modules[net.__module__].__doc__
            #         if desc is not None:
            #             desc = desc.strip("\n")

            #     table.add_row(model_name, desc)

            # console.print(table)

    else:
        console.print(
            Columns(
                model_list,
                padding=(0, 3),
                equal=True,
                column_first=True,
                title=f"[bold]{len(model_list)} Models[/bold]",
            )
        )



"""
ResNeSt, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/resnest.py

Paper "ResNeSt: Split-Attention Networks", https://arxiv.org/abs/2004.08955
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import make_divisible


class RadixSoftmax(nn.Module):
    def __init__(self, radix: int, cardinality: int) -> None:
        super().__init__()
        self.radix = radix
        self.cardinality = cardinality

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        B = x.size(0)
        if self.radix > 1:
            x = x.view(B, self.cardinality, self.radix, -1).transpose(1, 2)
            x = F.softmax(x, dim=1)
            x = x.reshape(B, -1)

        else:
            x = torch.sigmoid(x)

        return x


class SplitAttn(nn.Module):
    """
    Split-Attention (aka Splat)
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        groups: int,
        bias: bool,
        radix: int = 2,
        rd_ratio: float = 0.25,
        rd_divisor: int = 8,
    ) -> None:
        super().__init__()
        self.radix = radix
        mid_chs = out_channels * radix
        attn_chs = make_divisible(in_channels * radix * rd_ratio, min_value=32, divisor=rd_divisor)

        self.conv = Conv2dNormActivation(
            in_channels,
            mid_chs,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            groups=groups * radix,
            bias=bias,
        )

        self.fc1 = Conv2dNormActivation(
            out_channels, attn_chs, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), groups=groups
        )
        self.fc2 = nn.Conv2d(attn_chs, mid_chs, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), groups=groups)
        self.r_softmax = RadixSoftmax(radix, groups)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)

        (B, RC, H, W) = x.size()  # pylint: disable=invalid-name
        if self.radix > 1:
            x = x.reshape((B, self.radix, RC // self.radix, H, W))
            x_gap = x.sum(dim=1)
        else:
            x_gap = x

        x_gap = x_gap.mean((2, 3), keepdim=True)
        x_gap = self.fc1(x_gap)
        x_attn = self.fc2(x_gap)

        x_attn = self.r_softmax(x_attn).view(B, -1, 1, 1)
        if self.radix > 1:
            out = (x * x_attn.reshape((B, self.radix, RC // self.radix, 1, 1))).sum(dim=1)
        else:
            out = x * x_attn

        return out.contiguous()


class ResNeStBottleneck(nn.Module):
    def __init__(
        self, in_channels: int, out_channels: int, stride: tuple[int, int], expansion: int, radix: int
    ) -> None:
        super().__init__()
        assert radix >= 1

        if stride[0] > 1 or stride[1] > 1:
            avd_stride = stride
            stride = (1, 1)

        else:
            avd_stride = (0, 0)

        self.conv_norm_act1 = Conv2dNormActivation(
            in_channels,
            out_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            bias=False,
        )

        self.splat = SplitAttn(
            out_channels,
            out_channels,
            kernel_size=(3, 3),
            stride=stride,
            padding=(1, 1),
            groups=1,
            bias=False,
            radix=radix,
        )
        if avd_stride[0] > 0 or avd_stride[1] > 0:
            self.avd_last = nn.AvgPool2d(kernel_size=(3, 3), stride=avd_stride, padding=(1, 1))

        else:
            self.avd_last = nn.Identity()

        self.conv_norm_act2 = Conv2dNormActivation(
            out_channels,
            out_channels * expansion,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            bias=False,
            activation_layer=None,
        )

        if in_channels == out_channels * expansion:
            self.downsample = nn.Identity()

        else:
            if avd_stride == (0, 0):
                self.downsample = Conv2dNormActivation(
                    in_channels,
                    out_channels * expansion,
                    kernel_size=(1, 1),
                    stride=(1, 1),
                    padding=(0, 0),
                    bias=False,
                    activation_layer=None,
                )

            else:
                self.downsample = nn.Sequential(
                    nn.AvgPool2d(kernel_size=(2, 2), stride=avd_stride, padding=(0, 0), count_include_pad=False),
                    Conv2dNormActivation(
                        in_channels,
                        out_channels * expansion,
                        kernel_size=(1, 1),
                        stride=(1, 1),
                        padding=(0, 0),
                        bias=False,
                        activation_layer=None,
                    ),
                )

        self.relu = nn.ReLU(inplace=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        x = self.conv_norm_act1(x)
        x = self.splat(x)
        x = self.avd_last(x)
        x = self.conv_norm_act2(x)

        identity = self.downsample(identity)

        x += identity
        x = self.relu(x)

        return x


class ResNeSt(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        radix = 2
        expansion = 4
        filter_list = [64, 128, 256, 512]
        stem_width: int = self.config["stem_width"]
        units: list[int] = self.config["units"]
        final_drop: float = self.config["final_drop"]

        in_channels = stem_width * 2
        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels, stem_width, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False
            ),
            Conv2dNormActivation(stem_width, stem_width, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False),
            Conv2dNormActivation(
                stem_width, in_channels, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False
            ),
            nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i, (channels, num_blocks) in enumerate(zip(filter_list, units)):
            layers = []
            if i == 0:
                stride = (1, 1)
            else:
                stride = (2, 2)

            for block_idx in range(num_blocks):
                if block_idx != 0:
                    stride = (1, 1)

                layers.append(ResNeStBottleneck(in_channels, channels, stride, expansion, radix))
                in_channels = channels * expansion

            stages[f"stage{i+1}"] = nn.Sequential(*layers)
            return_channels.append(channels * expansion)

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
            nn.Dropout(p=final_drop),
        )
        self.return_channels = return_channels
        self.embedding_size = filter_list[-1] * expansion
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "resnest_14", ResNeSt, config={"stem_width": 32, "units": [1, 1, 1, 1], "final_drop": 0.0}
)
registry.register_model_config(
    "resnest_26", ResNeSt, config={"stem_width": 32, "units": [2, 2, 2, 2], "final_drop": 0.0}
)
registry.register_model_config(
    "resnest_50", ResNeSt, config={"stem_width": 32, "units": [3, 4, 6, 3], "final_drop": 0.0}
)
registry.register_model_config(
    "resnest_101", ResNeSt, config={"stem_width": 64, "units": [3, 4, 23, 3], "final_drop": 0.0}
)
registry.register_model_config(
    "resnest_200", ResNeSt, config={"stem_width": 64, "units": [3, 24, 36, 3], "final_drop": 0.2}
)
registry.register_model_config(
    "resnest_269", ResNeSt, config={"stem_width": 64, "units": [3, 30, 48, 8], "final_drop": 0.2}
)

registry.register_weights(
    "resnest_14_il-common",
    {
        "description": "ResNeSt 14 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 35.7,
                "sha256": "d80c2d290bdab79b03496853df87cb48fc01f63690500ab7d15086a2149ff1ba",
            }
        },
        "net": {"network": "resnest_14", "tag": "il-common"},
    },
)



import argparse
import logging
import multiprocessing as mp
import time
from typing import Any

import polars as pl
import torch
import torch.amp

import birder
from birder.common import cli
from birder.conf import settings
from birder.model_registry import registry

logger = logging.getLogger(__name__)


def dummy(arg: Any) -> None:
    type(arg)


def throughput_benchmark(
    net: torch.nn.Module, device: torch.device, sample_shape: tuple[int, ...], model_name: str, args: argparse.Namespace
) -> tuple[float, int]:
    # Sanity
    logger.info(f"Sanity check for {model_name}")

    if args.amp_dtype is None:
        amp_dtype = torch.get_autocast_dtype(device.type)
    else:
        amp_dtype = getattr(torch, args.amp_dtype)

    batch_size = sample_shape[0]
    while batch_size > 0:
        with torch.inference_mode():
            with torch.amp.autocast(device.type, enabled=args.amp, dtype=amp_dtype):
                try:
                    output = net(torch.rand(sample_shape, device=device))
                    output = net(torch.rand(sample_shape, device=device))
                    break
                except Exception:  # pylint: disable=broad-exception-caught
                    batch_size -= 32
                    sample_shape = (batch_size, *sample_shape[1:])
                    logger.info(f"Reducing batch size to: {batch_size}")

    if batch_size <= 0:
        logger.warning(f"Aborting {model_name} !!!")
        return (-1.0, 0)

    # Warmup
    logger.info(f"Starting warmup for {model_name}")
    with torch.inference_mode():
        with torch.amp.autocast(device.type, enabled=args.amp, dtype=amp_dtype):
            for _ in range(args.warmup):
                output = net(torch.rand(sample_shape, device=device))

    # Benchmark
    logger.info(f"Starting benchmark for {model_name}")
    with torch.inference_mode():
        with torch.amp.autocast(device.type, enabled=args.amp, dtype=amp_dtype):
            if device.type == "cuda":
                torch.cuda.synchronize(device=device)

            t_start = time.perf_counter()
            for _ in range(args.repeats):
                for _ in range(args.bench_iter):
                    output = net(torch.rand(sample_shape, device=device))

            if device.type == "cuda":
                torch.cuda.synchronize(device=device)

            t_end = time.perf_counter()
            t_elapsed = t_end - t_start

    dummy(output)

    return (t_elapsed, batch_size)


def memory_benchmark(
    sync_peak_memory: Any, sample_shape: tuple[int, ...], model_name: str, args: argparse.Namespace
) -> None:
    logger.info(f"Starting memory benchmark for {model_name}")
    if args.gpu is True:
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    if args.gpu_id is not None:
        torch.cuda.set_device(args.gpu_id)

    if args.amp_dtype is None:
        amp_dtype = torch.get_autocast_dtype(device.type)
    else:
        amp_dtype = getattr(torch, args.amp_dtype)

    (net, _) = birder.load_pretrained_model(model_name, inference=True, device=device)

    torch.cuda.empty_cache()
    torch.cuda.reset_peak_memory_stats(device)
    with torch.inference_mode():
        with torch.amp.autocast(device.type, enabled=args.amp, dtype=amp_dtype):
            sample = torch.rand(sample_shape, device=device)
            for _ in range(5):
                net(sample)

    peak_memory: float = torch.cuda.max_memory_allocated(device)
    sync_peak_memory.value = peak_memory


# pylint: disable=too-many-branches,too-many-locals
def benchmark(args: argparse.Namespace) -> None:
    mp.set_start_method("spawn")

    torch_version = torch.__version__
    output_path = "benchmark"
    if args.suffix is not None:
        output_path = f"{output_path}_{args.suffix}"

    benchmark_path = settings.RESULTS_DIR.joinpath(f"{output_path}.csv")
    if benchmark_path.exists() is True and args.append is False:  # pylint: disable=no-else-raise
        logger.warning("Benchmark file already exists... aborting")
        raise SystemExit(1)
    elif benchmark_path.exists() is True:
        logger.info(f"Loading {benchmark_path}...")
        existing_df = pl.read_csv(benchmark_path)
    else:
        existing_df = None

    if args.gpu is True:
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    if args.gpu_id is not None:
        torch.cuda.set_device(args.gpu_id)

    logger.info(f"Using device {device}")

    if args.fast_matmul is True or args.amp is True:
        torch.set_float32_matmul_precision("high")

    if args.single_thread is True:
        torch.set_num_threads(1)
        torch.set_num_interop_threads(1)

    input_channels = 3
    results = []
    model_list = birder.list_pretrained_models(args.filter)
    for model_name in model_list:
        model_metadata = registry.get_pretrained_metadata(model_name)
        size = model_metadata["resolution"]

        # Check if model already benchmarked at this configuration
        if existing_df is not None:
            combination_exists = existing_df.filter(
                **{
                    "model_name": model_name,
                    "device": device.type,
                    "single_thread": args.single_thread,
                    "compile": args.compile,
                    "amp": args.amp,
                    "fast_matmul": args.fast_matmul,
                    "size": size[0],
                    "max_batch_size": args.max_batch_size,
                    "memory": args.memory,
                }
            ).is_empty()
            if combination_exists is False:
                logger.info(f"{model_name} at the current configuration is already on file, skipping...")
                continue

        sample_shape = (args.max_batch_size, input_channels) + size

        if args.memory is True:
            samples_per_sec = None
            sync_peak_memory = mp.Value("d", 0.0)
            p = mp.Process(target=memory_benchmark, args=(sync_peak_memory, sample_shape, model_name, args))
            p.start()
            p.join()
            peak_memory = sync_peak_memory.value / (1024 * 1024)
            logger.info(f"{model_name} used {peak_memory:.2f}MB")
        else:
            # Initialize model
            (net, _) = birder.load_pretrained_model(model_name, inference=True, device=device)
            if args.compile is True:
                torch.compiler.reset()
                net = torch.compile(net)

            peak_memory = None
            (t_elapsed, batch_size) = throughput_benchmark(net, device, sample_shape, model_name, args)
            if t_elapsed < 0.0:
                continue

            num_samples = args.repeats * args.bench_iter * batch_size
            samples_per_sec = num_samples / t_elapsed
            logger.info(f"{model_name} ran at {samples_per_sec:.2f} samples / sec")

        results.append(
            {
                "model_name": model_name,
                "device": device.type,
                "single_thread": args.single_thread,
                "compile": args.compile,
                "amp": args.amp,
                "fast_matmul": args.fast_matmul,
                "size": size[0],
                "max_batch_size": args.max_batch_size,
                "memory": args.memory,
                "torch_version": torch_version,
                "samples_per_sec": samples_per_sec,
                "peak_memory": peak_memory,
            }
        )

    results_df = pl.DataFrame(results)

    if args.append is True and existing_df is not None:
        include_header = False
        mode = "a"
    else:
        include_header = True
        mode = "w"

    logger.info(f"Saving results at {benchmark_path}")
    with open(benchmark_path, mode=mode, encoding="utf-8") as handle:
        results_df.write_csv(handle, include_header=include_header)


def get_args_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="Benchmark pretrained models",
        epilog=(
            "Usage example:\n"
            "python benchmark.py --compile --suffix all\n"
            "python benchmark.py --filter '*il-common*' --compile --suffix il-common\n"
            "python benchmark.py --filter '*il-common*' --suffix il-common\n"
            "python benchmark.py --filter '*il-common*' --max-batch-size 512 --gpu\n"
            "python benchmark.py --filter '*il-common*' --max-batch-size 512 --gpu --warmup 20\n"
            "python benchmark.py --filter '*il-common*' --max-batch-size 512 --gpu --fast-matmul --compile "
            "--suffix il-common --append\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    parser.add_argument("--filter", type=str, help="models to benchmark (fnmatch type filter)")
    parser.add_argument("--compile", default=False, action="store_true", help="enable compilation")
    parser.add_argument(
        "--amp", default=False, action="store_true", help="use torch.amp.autocast for mixed precision inference"
    )
    parser.add_argument(
        "--amp-dtype",
        type=str,
        choices=["float16", "bfloat16"],
        help="whether to use float16 or bfloat16 for mixed precision",
    )
    parser.add_argument(
        "--fast-matmul", default=False, action="store_true", help="use fast matrix multiplication (affects precision)"
    )
    parser.add_argument("--max-batch-size", type=int, default=1, metavar="N", help="the max batch size to try")
    parser.add_argument("--suffix", type=str, help="add suffix to output file")
    parser.add_argument("--single-thread", default=False, action="store_true", help="use CPU with a single thread")
    parser.add_argument("--gpu", default=False, action="store_true", help="use gpu")
    parser.add_argument("--gpu-id", type=int, metavar="ID", help="gpu id to use")
    parser.add_argument("--warmup", type=int, default=20, metavar="N", help="number of warmup iterations")
    parser.add_argument("--repeats", type=int, default=3, metavar="N", help="number of repetitions")
    parser.add_argument("--bench-iter", type=int, default=300, metavar="N", help="number of benchmark iterations")
    parser.add_argument("--memory", default=False, action="store_true", help="benchmark memory instead of throughput")
    parser.add_argument("--append", default=False, action="store_true", help="append to existing output file")

    return parser


def validate_args(args: argparse.Namespace) -> None:
    assert args.single_thread is False or args.gpu is False
    assert args.memory is False or args.gpu is True
    assert args.memory is False or args.compile is False


def args_from_dict(**kwargs: Any) -> argparse.Namespace:
    parser = get_args_parser()
    parser.set_defaults(**kwargs)
    args = parser.parse_args([])
    validate_args(args)

    return args


def main() -> None:
    parser = get_args_parser()
    args = parser.parse_args()
    validate_args(args)

    if settings.RESULTS_DIR.exists() is False:
        logger.info(f"Creating {settings.RESULTS_DIR} directory...")
        settings.RESULTS_DIR.mkdir(parents=True)

    benchmark(args)


if __name__ == "__main__":
    logger = logging.getLogger(getattr(__spec__, "name", __name__))
    main()
